#!/usr/bin/env python
#
# orthotoolbar.py - The OrthoToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoToolBar` class, which is a
:class:`.ControlToolBar` for use with the :class:`.OrthoPanel`.
"""


import wx

import fsleyes_props                 as props
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.views.orthopanel      as orthopanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips
import fsleyes.actions               as actions
import fsleyes.strings               as strings


class OrthoToolBar(ctrlpanel.ControlToolBar):
    """The ``OrthoToolBar`` is a :class:`.ControlToolBar` for use with the
    :class:`.OrthoPanel`. An ``OrthoToolBar`` looks something like this:


    .. image:: images/orthotoolbar.png
       :scale: 50%
       :align: center


    The ``OrthoToolBar`` allows the user to control important parts of the
    :class:`.OrthoPanel` display, and also to display a
    :class:`.CanvasSettingsPanel`, which allows control over all aspects of
    an ``OrthoPanel``.

    The ``OrthoToolBar`` contains controls which modify properties, or run
    actions, defined on the following classes:

    .. autosummary::
       :nosignatures:

       ~fsleyes.views.orthopanel.OrthoPanel
       ~fsleyes.displaycontext.orthoopts.OrthoOpts
       ~fsleyes.profiles.orthoviewprofile.OrthoViewProfile
    """


    showCursorAndLabels = props.Boolean(default=True)
    """This property is linked to a button on the toolbar which allows the
    user to simultaneously toggle the :attr:`.SceneOpts.showCursor` and
    :attr:`.OrthoOpts.showLabels` properties.
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OrthoToolBar`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    def __init__(self, parent, overlayList, displayCtx, ortho):
        """Create an ``OrthoToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          ortho,
                                          height=24,
                                          kbFocus=True)

        self.orthoPanel = ortho

        # The toolbar has buttons bound to some actions
        # on the Profile  instance - when the profile
        # changes (between 'view' and 'edit'), the
        # Profile instance changes too, so we need
        # to re-create these action buttons. I'm being
        # lazy and just re-generating the entire toolbar.
        ortho.events.register(self.name, self.__makeTools, 'profile')

        self.addListener('showCursorAndLabels',
                         self.name,
                         self.__showCursorAndLabelsChanged)

        self.__makeTools()


    def destroy(self):
        """Must be called when this ``OrthoToolBar`` is no longer in use.
        Removes some property listeners, and calls the base class
        implementation.
        """
        self.orthoPanel.events.deregister(self.name, 'profile')
        self           .removeListener('showCursorAndLabels', self.name)

        ctrlpanel.ControlToolBar.destroy(self)


    def __makeTools(self, *a):
        """Called by :meth:`__init__`, and whenever the
        :attr:`.ViewPanel.profile` property changes.

        Re-creates all tools shown on this ``OrthoToolBar``.
        """

        ortho     = self.orthoPanel
        orthoOpts = ortho.sceneOpts
        profile   = ortho.currentProfile

        coronalIcon          = 'coronalSlice24'
        coronalHighlightIcon = 'coronalSliceHighlight24'

        icons = {
            'screenshot'       : fslicons.findImageFile('camera24'),
            'resetDisplay'     : fslicons.findImageFile('resetZoom24'),
            'showCursorAndLabels' : [
                fslicons.findImageFile('addHighlight24'),
                fslicons.findImageFile('add24')],
            'movieMode'        : [
                fslicons.findImageFile('movieHighlight24'),
                fslicons.findImageFile('movie24')],
            'showXCanvas'      : [
                fslicons.findImageFile('sagittalSliceHighlight24'),
                fslicons.findImageFile('sagittalSlice24')],
            'showYCanvas'      : [
                fslicons.findImageFile(coronalHighlightIcon),
                fslicons.findImageFile(coronalIcon)],
            'showZCanvas'      : [
                fslicons.findImageFile('axialSliceHighlight24'),
                fslicons.findImageFile('axialSlice24')],
            'CanvasSettingsPanel' : [
                fslicons.findImageFile('spannerHighlight24'),
                fslicons.findImageFile('spanner24')],

            'layout' : {
                'horizontal' : [
                    fslicons.findImageFile('horizontalLayoutHighlight24'),
                    fslicons.findImageFile('horizontalLayout24')],
                'vertical'   : [
                    fslicons.findImageFile('verticalLayoutHighlight24'),
                    fslicons.findImageFile('verticalLayout24')],
                'grid'       : [
                    fslicons.findImageFile('gridLayoutHighlight24'),
                    fslicons.findImageFile('gridLayout24')]}
        }

        tooltips = {
            'screenshot'   : fsltooltips.actions[   ortho,     'screenshot'],
            'resetDisplay' : fsltooltips.actions[   profile,   'resetDisplay'],
            'movieMode'    : fsltooltips.properties[ortho,     'movieMode'],
            'showCursorAndLabels' : fsltooltips.properties[
                self, 'showCursorAndLabels'],
            'zoom'         : fsltooltips.properties[orthoOpts, 'zoom'],
            'layout'       : fsltooltips.properties[orthoOpts, 'layout'],
            'showXCanvas'  : fsltooltips.properties[orthoOpts, 'showXCanvas'],
            'showYCanvas'  : fsltooltips.properties[orthoOpts, 'showYCanvas'],
            'showZCanvas'  : fsltooltips.properties[orthoOpts, 'showZCanvas'],
            'CanvasSettingsPanel' : fsltooltips.actions[ortho,
                                                        'CanvasSettingsPanel'],

        }

        targets    = {'screenshot'          : ortho,
                      'movieMode'           : ortho,
                      'showCursorAndLabels' : self,
                      'resetDisplay'        : profile,
                      'zoom'                : orthoOpts,
                      'layout'              : orthoOpts,
                      'showXCanvas'         : orthoOpts,
                      'showYCanvas'         : orthoOpts,
                      'showZCanvas'         : orthoOpts,
                      'CanvasSettingsPanel' : ortho}

        toolSpecs = [

            actions.ToggleActionButton(
                'CanvasSettingsPanel',
                actionKwargs={'floatPane' : True},
                icon=icons['CanvasSettingsPanel'],
                tooltip=tooltips['CanvasSettingsPanel']),
            actions.ActionButton('screenshot',
                                 icon=icons['screenshot'],
                                 tooltip=tooltips['screenshot']),
            'div',
            props  .Widget(      'showXCanvas',
                                 icon=icons['showXCanvas'],
                                 tooltip=tooltips['showXCanvas']),
            props  .Widget(      'showYCanvas',
                                 icon=icons['showYCanvas'],
                                 tooltip=tooltips['showYCanvas']),
            props  .Widget(      'showZCanvas',
                                 icon=icons['showZCanvas'],
                                 tooltip=tooltips['showZCanvas']),
            'div',
            props  .Widget(      'layout',
                                 icons=icons['layout'],
                                 tooltip=tooltips['layout']),
            'div',
            props  .Widget(      'movieMode',
                                 icon=icons['movieMode'],
                                 tooltip=tooltips['movieMode']),
            props  .Widget(      'showCursorAndLabels',
                                 icon=icons['showCursorAndLabels'],
                                 tooltip=tooltips['showCursorAndLabels']),
            actions.ActionButton('resetDisplay',
                                 icon=icons['resetDisplay'],
                                 tooltip=tooltips['resetDisplay']),
            props.Widget(        'zoom',
                                 spin=True,
                                 slider=True,
                                 showLimits=False,
                                 spinWidth=5,
                                 tooltip=tooltips['zoom']),
        ]

        tools = []
        nav   = []

        for spec in toolSpecs:

            if spec == 'div':
                tools.append(fsltoolbar.ToolBarDivider(self,
                                                       height=24,
                                                       orient=wx.VERTICAL))
                continue

            widget    = props.buildGUI(self, targets[spec.key], spec)
            navWidget = widget

            if spec.key in ('zoom', ):
                widget = self.MakeLabelledTool(
                    widget,
                    strings.properties[targets[spec.key], spec.key])

            tools.append(widget)
            nav  .append(navWidget)

        self.SetTools(tools, destroy=True)
        self.setNavOrder(nav)


    def __showCursorAndLabelsChanged(self, *a):
        """Called when the :attr:`showCursorAndLabels` property is changed.
        Propagates the change on to the :attr:`.SceneOpts.showCursor` and
        :attr:`.OrthoOpts.showLabels` properties.
        """

        opts            = self.orthoPanel.sceneOpts
        opts.showCursor = self.showCursorAndLabels
        opts.showLabels = self.showCursorAndLabels



#!/usr/bin/env python
#
# colourbar.py - The ColourBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ColourBar` class, which generates a bitmap
rendering of a colour bar.
"""


import numpy as np

import fsl.utils.notifier                    as notifier
import fsleyes_props                         as props
import fsleyes_widgets.utils.colourbarbitmap as cbarbmp

import fsleyes.displaycontext                as fsldc
import fsleyes.displaycontext.colourmapopts  as cmapopts


def colourBarMinorAxisSize(fontSize):
    """Calculates a good size for the minor axis of a colour bar.

    The minor axis is the axis perpendicular to the colour bar axis.

    :arg fontSize: Font size of colour bar labels, in points.
    """

    # Figure out the font size in pixels
    # (font points are 1/72th of an inch,
    # and we're using inside knowledge
    # that the colourbarbitmap module
    # uses 96 dpi, and a padding of 6
    # pixels).
    fontSize = fontSize
    fontSize = 6 + 96 * fontSize / 72.

    # Fix the minor axis of the colour bar,
    # according to the font size, and a
    # constant size for the colour bar
    return round(2 * fontSize + 40)


class ColourBar(props.HasProperties, notifier.Notifier):
    """A ``ColourBar`` is an object which listens to the properties of a
    :class:`.ColourMapOpts` instance, and automatically generates a colour
    bar bitmap representing the current colour map properties.

    Whenever the colour bar is refreshed, a notification is emitted via the
    :class:`.Notifier` interface.
    """


    orientation = props.Choice(('horizontal', 'vertical'))
    """Whether the colour bar should be vertical or horizontal. """


    labelSide = props.Choice(('top-left', 'bottom-right'))
    """Whether the colour bar labels should be on the top/left, or bottom/right
    of the colour bar (depending upon whether the colour bar orientation is
    horizontal/vertical).
    """


    textColour = props.Colour(default=(1, 1, 1, 1))
    """Colour to use for the colour bar label. """


    bgColour = props.Colour(default=(0, 0, 0, 1))
    """Colour to use for the background. """


    showLabel = props.Boolean(default=True)
    """Toggle the colour bar label (the :attr:`.Display.name` property). """


    showTicks = props.Boolean(default=True)
    """Toggle the tick labels (the :attr:`.ColourMapOpts.displayRange`). """


    fontSize = props.Int(minval=4, maxval=96, default=12)
    """Size of the font used for the text on the colour bar."""


    def __init__(self, overlayList, displayCtx):
        """Create a ``ColourBar``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        """


        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__name        = '{}_{}'.format(type(self).__name__, id(self))

        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)

        self.addGlobalListener(self.name, self.__clearColourBar)

        self.__opts      = None
        self.__display   = None
        self.__size      = (None, None, None)
        self.__colourBar = None

        self.__selectedOverlayChanged()


    @property
    def name(self):
        """Return the name of this ColourBar, used internally for registering
        property listeners.
        """
        return self.__name


    def destroy(self):
        """Must be called when this ``ColourBar`` is no longer needed.

        Removes all registered listeners from the :class:`.OverlayList`,
        :class:`.DisplayContext`, and foom individual overlays.
        """


        self.__overlayList.removeListener('overlays',        self.name)
        self.__displayCtx .removeListener('selectedOverlay', self.name)
        self.__deregisterOverlay()


    def __selectedOverlayChanged(self, *a):
        """Called when the :class:`.OverlayList` or the
        :attr:`.DisplayContext.selectedOverlay` changes.

        If the newly selected overlay is being displayed with a
        :class:`.ColourMapOpts` instance, various property listeners are
        registered, and the colour bar is refreshed.
        """

        self.__deregisterOverlay()
        self.__registerOverlay()
        self.__clearColourBar()


    def __deregisterOverlay(self):
        """Called when the selected overlay changes. De-registers property
        listeners from any previously-registered :class:`.ColourMapOpts`
        instance.
        """

        if self.__opts is None:
            return

        try:
            opts    = self.__opts
            display = self.__display

            opts   .removeListener('displayRange',    self.name)
            opts   .removeListener('cmap',            self.name)
            opts   .removeListener('negativeCmap',    self.name)
            opts   .removeListener('useNegativeCmap', self.name)
            opts   .removeListener('invert',          self.name)
            opts   .removeListener('gamma',           self.name)
            opts   .removeListener('cmapResolution',  self.name)
            display.removeListener('name',            self.name)

        except fsldc.InvalidOverlayError:
            pass

        self.__opts    = None
        self.__display = None


    def __registerOverlay(self):
        """Called when the selected overlay changes. Registers property
        listeners with the :class:`.ColourMapOpts` instance associated with
        the newly selected overlay.
        """

        overlay = self.__displayCtx.getSelectedOverlay()

        if overlay is None:
            return False

        display = self.__displayCtx.getDisplay(overlay)
        opts    = display.opts

        if not isinstance(opts, cmapopts.ColourMapOpts):
            return False

        self.__opts    = opts
        self.__display = display

        opts   .addListener('displayRange',
                            self.name,
                            self.__clearColourBar)
        opts   .addListener('cmap',
                            self.name,
                            self.__clearColourBar)
        opts   .addListener('negativeCmap',
                            self.name,
                            self.__clearColourBar)
        opts   .addListener('useNegativeCmap',
                            self.name,
                            self.__clearColourBar)
        opts   .addListener('invert',
                            self.name,
                            self.__clearColourBar)
        opts   .addListener('cmapResolution',
                            self.name,
                            self.__clearColourBar)
        opts   .addListener('gamma',
                            self.name,
                            self.__clearColourBar)
        display.addListener('name',
                            self.name,
                            self.__clearColourBar)

        return True


    def __clearColourBar(self, *a):
        """Clears any previously generated colour bar bitmap. """
        self.__colourBar = None
        self.notify()


    def colourBar(self, w, h, scale=1):
        """Returns a bitmap containing the rendered colour bar, rendering it if
        necessary.

        :arg w:     Width in pixels
        :arg h:     Height in pixels
        :arg scale: DPI scaling factor, if applicable.
        """

        if self.__opts is None:
            return None

        if w < 20: w = 20
        if h < 20: h = 20

        if (w, h, scale) == self.__size and self.__colourBar is not None:
            return self.__colourBar

        display        = self.__display
        opts           = self.__opts
        cmap           = opts.cmap
        negCmap        = opts.negativeCmap
        useNegCmap     = opts.useNegativeCmap
        cmapResolution = opts.cmapResolution
        gamma          = opts.realGamma(opts.gamma)
        invert         = opts.invert
        dmin, dmax     = opts.displayRange.x
        label          = display.name

        if self.orientation == 'horizontal':
            if  self.labelSide == 'top-left': labelSide = 'top'
            else:                             labelSide = 'bottom'
        else:
            if  self.labelSide == 'top-left': labelSide = 'left'
            else:                             labelSide = 'right'

        if useNegCmap and dmin == 0.0:
            ticks      = [0.0, 0.5, 1.0]
            ticklabels = ['{:0.3G}'.format(-dmax),
                          '{:0.3G}'.format( dmin),
                          '{:0.3G}'.format( dmax)]
            tickalign  = ['left', 'center', 'right']
        elif useNegCmap:
            ticks      = [0.0, 0.49, 0.51, 1.0]
            ticklabels = ['{:0.3G}'.format(-dmax),
                          '{:0.3G}'.format(-dmin),
                          '{:0.3G}'.format( dmin),
                          '{:0.3G}'.format( dmax)]
            tickalign  = ['left', 'right', 'left', 'right']
        else:
            negCmap    = None
            ticks      = [0.0, 1.0]
            tickalign  = ['left', 'right']
            ticklabels = ['{:0.3G}'.format(dmin),
                          '{:0.3G}'.format(dmax)]

        ticks = np.array(ticks)
        ticks[np.isclose(ticks , 0)] = 0

        if not self.showLabel:
            label = None
        if not self.showTicks:
            ticks      = None
            ticklabels = None

        bitmap = cbarbmp.colourBarBitmap(
            cmap=cmap,
            negCmap=negCmap,
            invert=invert,
            gamma=gamma,
            ticks=ticks,
            ticklabels=ticklabels,
            tickalign=tickalign,
            width=w,
            height=h,
            label=label,
            scale=scale,
            orientation=self.orientation,
            labelside=labelSide,
            textColour=self.textColour,
            fontsize=self.fontSize,
            bgColour=self.bgColour,
            cmapResolution=cmapResolution)

        self.__size      = (w, h, scale)
        self.__colourBar = bitmap

        return bitmap



#!/usr/bin/env python
#
# sceneopts.py - Provides the SceneOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.SceneOpts` class, which contains display
settings used by :class:`.CanvasPanel` instances.
"""


import copy
import logging

import fsleyes_props      as props
import fsleyes.gl         as fslgl
import fsleyes.colourmaps as fslcm

from . import canvasopts


log = logging.getLogger(__name__)


class SceneOpts(props.HasProperties):
    """The ``SceneOpts`` class defines settings which are used by
    :class:`.CanvasPanel` instances.

    Several of the properties of the ``SceneOpts`` class are defined in the
    :class:`.SliceCanvasOpts` class, so see its documentation for more
    details.
    """


    showCursor      = copy.copy(canvasopts.SliceCanvasOpts.showCursor)
    zoom            = copy.copy(canvasopts.SliceCanvasOpts.zoom)
    bgColour        = copy.copy(canvasopts.SliceCanvasOpts.bgColour)
    cursorColour    = copy.copy(canvasopts.SliceCanvasOpts.cursorColour)
    renderMode      = copy.copy(canvasopts.SliceCanvasOpts.renderMode)
    highDpi         = copy.copy(canvasopts.SliceCanvasOpts.highDpi)


    fgColour = props.Colour(default=(1, 1, 1))
    """Colour to use for foreground items (e.g. labels).

    .. note:: This colour is automatically updated whenever the
              :attr:`.bgColour` is changed. But it can be modified
              independently.
    """


    showColourBar = props.Boolean(default=False)
    """If ``True``, and it is possible to do so, a colour bar is shown on
    the scene.
    """


    colourBarLocation  = props.Choice(('top', 'bottom', 'left', 'right'))
    """This property controls the location of the colour bar, if it is being
    shown.
    """


    colourBarLabelSide = props.Choice(('top-left', 'bottom-right'))
    """This property controls the location of the colour bar labels, relative
    to the colour bar, if it is being shown.
    """


    colourBarSize = props.Percentage(default=100)
    """Size of the major axis of the colour bar, as a proportion of the
    available space.
    """


    labelSize = props.Int(minval=4, maxval=96, default=12, clamped=True)
    """Font size used for any labels drawn on the canvas, including
    orthographic labels, and colour bar labels.
    """


    # NOTE: If you change the maximum performance value,
    #       make sure you update all references to
    #       performance because, for example, the
    #       OrthoEditProfile does numerical comparisons
    #       to it.
    performance = props.Choice((1, 2, 3), default=3, allowStr=True)
    """User controllable performance setting.

    This property is linked to the :attr:`renderMode` property. Setting this
    property to a low value will result in faster rendering time, at the cost
    of increased memory usage and poorer rendering quality.

    See the :meth:`__onPerformanceChange` method.
    """


    movieSyncRefresh = props.Boolean(default=True)
    """Whether, when in movie mode, to synchronise the refresh for GL
    canvases. This is not possible in some platforms/environments. See
    :attr:`.CanvasPanel.movieSyncRefresh`.
    """


    def __init__(self, panel):
        """Create a ``SceneOpts`` instance.

        This method simply links the :attr:`performance` property to the
        :attr:`renderMode` property.
        """

        self.__panel = panel
        self.__name  = '{}_{}'.format(type(self).__name__, id(self))

        self.movieSyncRefresh = self.defaultMovieSyncRefresh

        self.addListener('performance', self.__name, self._onPerformanceChange)
        self.addListener('bgColour',    self.__name, self.__onBgColourChange)
        self._onPerformanceChange()


    @property
    def defaultMovieSyncRefresh(self):
        """In movie mode, the canvas refreshes are performed by the
        __syncMovieRefresh or __unsyncMovieRefresh methods of the CanvasPanel
        class. Some platforms/GL drivers/environments seem to have a problem
        with separate renders/buffer swaps, so we have to use a shitty
        unsynchronised update routine.

        These heuristics are not perfect - the movieSyncRefresh property can
        therefore be overridden by the user.
        """
        renderer        = fslgl.GL_RENDERER.lower()
        unsyncRenderers = ['gallium', 'mesa dri intel(r)']
        unsync          = any([r in renderer for r in unsyncRenderers])

        return not unsync


    @property
    def panel(self):
        """Return a reference to the ``CanvasPanel`` that owns this
        ``SceneOpts`` instance.
        """
        return self.__panel


    def _onPerformanceChange(self, *a):
        """Called when the :attr:`performance` property changes.

        This method must be overridden by sub-classes to change the values of
        the :attr:`renderMode` property according to the new performance
        setting.
        """
        raise NotImplementedError('The _onPerformanceChange method must'
                                  'be implemented by sub-classes')


    def __onBgColourChange(self, *a):
        """Called when the background colour changes. Updates the
        :attr:`fgColour` to a complementary colour.
        """
        self.fgColour = fslcm.complementaryColour(self.bgColour)



#!/usr/bin/env python
#
# browsexnat.py - Action which allows the user to connect to and browse an
# XNAT repository.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`BrowseXNATAction`, which allows the user
to connect to and browse an XNAT repository. If ``wxnatpy``
(https://github.com/pauldmccarthy/wxnatpy) is not present, the action is
disabled.
"""


import            os
import os.path as op

import wx

import fsl.utils.settings  as fslsettings
import fsleyes.strings     as strings
import fsleyes.autodisplay as autodisplay

from . import base
from . import loadoverlay

# if wxnatpy is not present, the
# action is permanently disabled
try:
    import wxnat
except ImportError:
    wxnat = None


class BrowseXNATAction(base.Action):
    """The ``BrowseXNATAction`` allows the user to open files from an XNAT
    repository. It opens a :class:`XNATBrowser``, and adds the files that
    the user selected into the :class:`.OverlayList`.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``BrowseXNATAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__openBrowser)
        self.__frame = frame

        if wxnat is None:
            self.enabled = False


    def __openBrowser(self):
        """Opens a :class:`XNATBrowser`, then adds any files that the user
        selected to the :class:`.OverlayList`.
        """

        hosts    = fslsettings.read('fsleyes.xnat.hosts',    [])
        accounts = fslsettings.read('fsleyes.xnat.accounts', {})

        dlg = XNATBrowser(self.__frame, hosts, accounts)

        dlg.Layout()
        dlg.Fit()
        dlg.SetSize((-1, 400))
        dlg.CentreOnParent()

        if dlg.ShowModal() != wx.ID_OK:
            return

        paths    = dlg.GetPaths()
        hosts    = dlg.GetHosts()
        accounts = dlg.GetAccounts()

        # No files downloaded
        if len(paths) == 0:
            return

        # Save successful hosts/credentials
        fslsettings.write('fsleyes.xnat.hosts',    hosts)
        fslsettings.write('fsleyes.xnat.accounts', accounts)

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)
            self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[-1]

            if self.displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.overlayList,
                                            self.displayCtx)

        loadoverlay.loadOverlays(paths,
                                 onLoad=onLoad,
                                 saveDir=False,
                                 inmem=self.displayCtx.loadInMemory)


class XNATBrowser(wx.Dialog):
    """The ``XNATBrowser`` contains a ``wxnat.XNATBrowserPanel``, allowing the
    user to connect to and browse an XNAT repository. It contains a *Download*
    button which, when clicked, downloads all selected files from the
    repository into a temporary directory. Once the files have been downloaded,
    their paths can be retrieved via the :meth:`GetPaths` method.
    """


    def __init__(self,
                 parent,
                 knownHosts=None,
                 knownAccounts=None):
        """Create a ``XNATBrowser``.

        :arg parent:        ``wx`` parent object

        :arg knownHosts:    List of hosts to use as auto-complete options

        :arg knownAccounts: Mapping containing login credentials, in the
                            ``{ host : (username, password) }``.
        """

        wx.Dialog.__init__(self,
                           parent,
                           title=strings.titles[self],
                           style=wx.RESIZE_BORDER)

        if wxnat is None:
            raise RuntimeError('wxnatpy is not available!')

        filters = {'file' : '*.nii|*.nii.gz|*.img|*.img.gz|*.gii|*.vtk'}

        self.__panel = wxnat.XNATBrowserPanel(
            self,
            knownHosts=knownHosts,
            knownAccounts=knownAccounts,
            filterType='glob',
            filters=filters)

        self.__ok       = wx.Button(self, wx.ID_OK)
        self.__cancel   = wx.Button(self, wx.ID_CANCEL)
        self.__btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer    = wx.BoxSizer(wx.VERTICAL)

        self.__ok    .SetLabel(strings.labels[self, 'ok'])
        self.__cancel.SetLabel(strings.labels[self, 'cancel'])

        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND, proportion=1)
        self.__btnSizer.Add(self.__ok,     flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND)
        self.__btnSizer.Add(self.__cancel, flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND)

        self.__sizer.Add((1, 10),         flag=wx.EXPAND)
        self.__sizer.Add(self.__panel,    flag=wx.EXPAND, proportion=1)
        self.__sizer.Add((1, 10),         flag=wx.EXPAND)
        self.__sizer.Add(self.__btnSizer, flag=wx.EXPAND)
        self.__sizer.Add((1, 10),         flag=wx.EXPAND)

        self.SetSizer(self.__sizer)
        self.__ok.SetDefault()

        self.__panel .Bind(wxnat.EVT_XNAT_FILE_SELECT_EVENT, self.__onOk)
        self.__ok    .Bind(wx.EVT_BUTTON,                    self.__onOk)
        self.__cancel.Bind(wx.EVT_BUTTON,                    self.__onCancel)

        self.__paths = []


    def GetPaths(self):
        """Returns paths to the files that were downloaded. """
        return self.__paths


    def GetHosts(self):
        """Wraps ``wxnat.XNATBrowserPanel.GetHosts``. """
        return self.__panel.GetHosts()


    def GetAccounts(self):
        """Wraps ``wxnat.XNATBrowserPanel.GetAccounts``. """
        return self.__panel.GetAccounts()


    def __onOk(self, ev):
        """Called when the *Ok* button is pushed. Prompts the user to select a
        directory, and then downloads the files.
        """

        files = self.__panel.GetSelectedFiles()

        if len(files) == 0:
            self.EndModal(wx.ID_OK)

        destdir = fslsettings.read('fsleyes.xnat.downloaddir', os.getcwd())
        dlg     = wx.DirDialog(self,
                               strings.labels[self, 'choosedir'],
                               defaultPath=destdir)

        if dlg.ShowModal() != wx.ID_OK:
            return

        destdir = dlg.GetPath()

        for f in files:
            dest = self.__panel.DownloadFile(f, op.join(destdir, f.id))

            if dest is not None:
                self.__paths.append(dest)

        fslsettings.write('fsleyes.xnat.downloaddir', destdir)

        self.__panel.EndSession()
        self.EndModal(wx.ID_OK)


    def __onCancel(self, ev):
        """Called when the *Cancel* button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_CANCEL)



#!/usr/bin/env python
#
# canvaspanel.py - Base class for all panels that display overlay data.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`CanvasPanel` class, which is the base
class for all panels which display overlays using ``OpenGL``.
"""


import logging

import wx

import fsl.utils.idle         as idle
import fsleyes_props          as props
import fsleyes.actions        as actions
import fsleyes.displaycontext as displayctx
from . import                    colourbarpanel
from . import                    viewpanel


log = logging.getLogger(__name__)


class CanvasPanel(viewpanel.ViewPanel):
    """The ``CanvasPanel`` class is a :class:`.ViewPanel` which is the base
    class for all panels which display overlays using ``OpenGL``
    (e.g. the :class:`.OrthoPanel` and the :class:`.LightBoxPanel`). A
    ``CanvasPanel`` instance uses a :class:`.SceneOpts` instance to control
    much of its functionality. The ``SceneOpts`` instance used by a
    ``CanvasPanel`` can be accessed via the :meth:`sceneOpts` property.


    The ``CanvasPanel`` class contains settings and functionality common to
    all sub-classes, including *movie mode* (see :attr:`movieMode`), the
    ability to show a colour bar (a :class:`.ColourBarPanel`; see
    :attr:`.SceneOpts.showColourBar`), and a number of actions.


    **Sub-class implementations**


    Sub-classes of the ``CanvasPanel`` must do the following:

      1. Add their content to the panel that is accessible via the
         :meth:`contentPanel` property (see the note on
         :ref:`adding content <canvaspanel-adding-content>`).

      2. Override the :meth:`getGLCanvases` method.

      3. Call the :meth:`centrePanelLayout` method in their ``__init__``
         method.

      4. Override the :meth:`centrePanelLayout` method if any custom layout is
         necessary.

      5. Call :meth:`.ViewPanel.initProfile` with the initial interaction
         profile


    **Actions**


    The following actions are available through a ``CanvasPanel`` (see
    the :mod:`.actions` module):

    .. autosummary::
       :nosignatures:

       screenshot
       movieGif
       showCommandLineArgs
       toggleMovieMode
       toggleDisplaySync
       toggleVolumeSync

    .. _canvaspanel-adding-content:


    **Adding content**


    To support colour bar and screenshot functionality, the ``CanvasPanel``
    uses a hierarchy of ``wx.Panel`` instances, depicted in the following
    containment hierarchy diagram:

    .. graphviz::

       digraph canvasPanel {

         graph [size=""];

         node [style="filled",
               shape="box",
               fillcolor="#ddffdd",
               fontname="sans"];

         rankdir="BT";

         1 [label="CanvasPanel"];
         2 [label="Centre panel"];
         3 [label="Custom content (for complex layouts)"];
         4 [label="Container panel"];
         5 [label="ColourBarPanel"];
         6 [label="Content panel"];
         7 [label="Content added by sub-classes"];

         2 -> 1;
         3 -> 2;
         4 -> 2;
         5 -> 4;
         6 -> 4;
         7 -> 6;
       }


    As depicted in the diagram, sub-classes need to add their content to the
    *content panel*. This panel is accessible via the :meth:`contentPanel`
    property.


    The *centre panel* is the :meth:`.ViewPanel.centrePanel`. The *container
    panel* is also available, via :meth:`containerPanel`. Everything in
    the container panel will appear in screenshots (see the :meth:`screenshot`
    method).


    The :meth:`centrePanelLayout` method lays out the centre panel, using the
    :meth:`layoutContainerPanel` method to lay out the colour bar and the
    content panel. The ``centrePanelLayout`` method simply adds the canvas
    container directly to the centre panel. Sub-classes which have more
    advanced layout requirements (e.g.  the :class:`.LightBoxPanel` needs a
    scrollbar) may override the :meth:`centrePanelLayout` method to implement
    their own layout.  These sub-class implementations must:

      1. Call the :meth:`layoutContainerPanel` method.

      2. Add the container panel (accessed via :meth:`containerPanel`)
         to the centre panel (accessed via :meth:`centrePanel`).

      3. Add any other custom content to the centre panel.
    """


    syncLocation = props.Boolean(default=True)
    """If ``True`` (the default), the :attr:`.DisplayContext.location` for
    this ``CanvasPanel`` is linked to the master ``DisplayContext`` location.
    """


    syncOverlayOrder = props.Boolean(default=True)
    """If ``True`` (the default), the :attr:`.DisplayContext.overlayOrder`
    for this ``CanvasPanel`` is linked to the master ``DisplayContext``
    overlay order.
    """


    syncOverlayDisplay = props.Boolean(default=True)
    """If ``True`` (the default), the properties of the :class:`.Display`
    and :class:`.DisplayOpts` instances for every overlay, as managed
    by the :attr:`.DisplayContext` for this ``CanvasPanel``, are linked to
    the properties of all ``Display`` and ``DisplayOpts`` instances managed
    by the master ``DisplayContext`` instance.
    """


    syncOverlayVolume = props.Boolean(default=True)
    """If ``True`` (the default), the volume/timepoint properties of the
    :class:`.DisplayOpts` instances for every overlay, as managed by the
    :attr:`.DisplayContext` for this ``CanvasPanel``, are linked to the
    properties of all ``DisplayOpts`` instances managed by the master
    ``DisplayContext`` instance.
    """


    movieMode = props.Boolean(default=False)
    """If ``True``, and the currently selected overlay (see
    :attr:`.DisplayContext.selectedOverlay`) is a :class:`.Image` instance
    with its display managed by a :class:`.VolumeOpts` instance, the displayed
    volume is changed periodically, according to the :attr:`movieRate`
    property.

    The update is performed on the main application thread via
    ``wx.CallLater``.
    """


    movieRate = props.Int(minval=10, maxval=500, default=400, clamped=True)
    """The movie update rate in milliseconds. The value of this property is
    inverted so that a high value corresponds to a fast rate, which makes
    more sense when displayed as an option to the user.
    """


    movieAxis = props.Choice((0, 1, 2, 3), default=3)
    """Axis along which the movie should be played, relative to the
    currently selected :class:`.Image`.
    """


    movieSyncRefresh = props.Boolean(default=True)
    """Whether, when in movie mode, to synchronise the refresh for GL
    canvases. This is not possible in some platforms/environments.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, sceneOpts):
        """Create a ``CanvasPanel``.

        :arg parent:       The :mod:`wx` parent object.

        :arg overlayList:  The :class:`.OverlayList` instance.

        :arg displayCtx:   The :class:`.DisplayContext` instance.

        :arb frame:        The :class:`.FSLeyesFrame` instance.

        :arg sceneOpts:    A :class:`.SceneOpts` instance for this
                           ``CanvasPanel`` - must be created by
                           sub-classes.
        """

        viewpanel.ViewPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__opts = sceneOpts

        # Use this name for listener registration,
        # in case subclasses use the FSLeyesPanel.name
        self.__name = 'CanvasPanel_{}'.format(self.name)

        # Bind the sync* properties of this
        # CanvasPanel to the corresponding
        # properties on the DisplayContext
        # instance.
        if displayCtx.getParent() is not None:
            self.bindProps('syncLocation',
                           displayCtx,
                           displayCtx.getSyncPropertyName('worldLocation'))
            self.bindProps('syncOverlayOrder',
                           displayCtx,
                           displayCtx.getSyncPropertyName('overlayOrder'))
            self.bindProps('syncOverlayDisplay', displayCtx)
            self.bindProps('syncOverlayVolume',  displayCtx)

        # If the displayCtx instance does not
        # have a parent, this means that it is
        # a top level instance
        else:
            self.disableProperty('syncLocation')
            self.disableProperty('syncOverlayOrder')
            self.disableProperty('syncOverlayDisplay')
            self.disableProperty('syncOverlayVolume')

        import fsleyes.actions.moviegif as moviegif

        self.centrePanel      = wx.Panel(self)
        self.__containerPanel = wx.Panel(self.centrePanel)
        self.__contentPanel   = wx.Panel(self.__containerPanel)
        self.__movieGifAction = moviegif.MovieGifAction(
            overlayList, displayCtx, self)

        self.bindProps('movieSyncRefresh', sceneOpts)

        self.toggleMovieMode  .bindProps('toggled', self, 'movieMode')
        self.toggleDisplaySync.bindProps('toggled', self, 'syncOverlayDisplay')
        self.toggleVolumeSync .bindProps('toggled', self, 'syncOverlayVolume')
        self.movieGif         .bindProps('enabled', self.__movieGifAction)

        # the __movieModeChanged method is called
        # when movieMode changes, but also when
        # the movie axis, overlay list, or selected
        # overlay changes. This is because, if movie
        # mode is on, but no overlay, or an
        # incompatible overlay, is selected, the
        # movie loop stops. So it needs to be
        # re-started if/when a compatible overlay is
        # selected.
        self.__movieRunning = False
        self            .addListener('movieMode',
                                     self.__name,
                                     self.__movieModeChanged)
        self            .addListener('movieAxis',
                                     self.__name,
                                     self.__movieModeChanged)
        self.overlayList.addListener('overlays',
                                     self.__name,
                                     self.__movieModeChanged)
        self.displayCtx .addListener('selectedOverlay',
                                     self.__name,
                                     self.__movieModeChanged)

        # Canvas/colour bar layout is managed
        # in the layoutContainerPanel method
        self.__colourBar = None

        self.__opts.addListener('colourBarLocation',
                                self.__name,
                                self.__colourBarPropsChanged)
        self.__opts.addListener('showColourBar',
                                self.__name,
                                self.__colourBarPropsChanged)
        self.__opts.addListener('bgColour',
                                self.__name,
                                self.__bgfgColourChanged)
        self.__opts.addListener('fgColour',
                                self.__name,
                                self.__bgfgColourChanged)
        self.__opts.addListener('labelSize',
                                self.__name,
                                self.__labelSizeChanged)


        idle.idle(self.__bgfgColourChanged)


    def destroy(self):
        """Makes sure that any remaining control panels are destroyed
        cleanly, and calls :meth:`.ViewPanel.destroy`.
        """

        if self.__colourBar is not None:
            self.__colourBar.destroy()

        self            .removeListener('movieMode',         self.__name)
        self            .removeListener('movieAxis',         self.__name)
        self.overlayList.removeListener('overlays',          self.__name)
        self.displayCtx .removeListener('selectedOverlay',   self.__name)
        self.sceneOpts  .removeListener('colourBarLocation', self.__name)
        self.sceneOpts  .removeListener('showColourBar',     self.__name)
        self.sceneOpts  .removeListener('bgColour',          self.__name)
        self.sceneOpts  .removeListener('fgColour',          self.__name)
        self.sceneOpts  .removeListener('labelSize',         self.__name)
        self.__movieGifAction.destroy()

        self.__opts           = None
        self.__movieGifAction = None

        viewpanel.ViewPanel.destroy(self)


    @actions.action
    def screenshot(self):
        """Takes a screenshot of the currently displayed scene on this
        ``CanvasPanel``.

        See the :class:`.ScreenshotAction`.
        """
        from fsleyes.actions.screenshot import ScreenshotAction
        ScreenshotAction(self.overlayList, self.displayCtx, self)()


    @actions.action
    def movieGif(self):
        """Generates an animated GIF of the currently displayed scene and
        movie mode settings on this ``CanvasPanel``.

        See the :class:`.MovieGifAction`.
        """
        self.__movieGifAction()


    @actions.action
    def showCommandLineArgs(self):
        """Shows the command line arguments which can be used to re-create
        the currently displayed scene. See the :class:`.ShowCommandLineAction`
        class.
        """
        from fsleyes.actions.showcommandline import ShowCommandLineAction
        ShowCommandLineAction(self.overlayList, self.displayCtx, self)()


    @actions.action
    def applyCommandLineArgs(self):
        """Shows the command line arguments which can be used to re-create
        the currently displayed scene. See the :class:`.ApplyCommandLineAction`
        class.
        """
        from fsleyes.actions.applycommandline import ApplyCommandLineAction
        ApplyCommandLineAction(self.overlayList, self.displayCtx, self)()


    @actions.toggleAction
    def toggleMovieMode(self):
        """Toggles the value of :attr:`movieMode`. """
        # The state of this action gets bound to
        # the movieMode attribute in __init__
        pass


    @actions.toggleAction
    def toggleDisplaySync(self):
        """Toggles the value of :attr:`syncOverlayDisplay`. """
        # The state of this action gets bound to
        # the syncOverlayDisplay attribute in __init__
        pass


    @actions.toggleAction
    def toggleVolumeSync(self):
        """Toggles the value of :attr:`syncOverlayVolume`. """
        # The state of this action gets bound to
        # the syncOverlayVolume attribute in __init__
        pass


    @property
    def sceneOpts(self):
        """Returns the :class:`.SceneOpts` instance used by this
        ``CanvasPanel``.
        """
        return self.__opts


    @property
    def contentPanel(self):
        """Returns the ``wx.Panel`` to which sub-classes must add their content.
        See the note on :ref:`adding content <canvaspanel-adding-content>`.
        """
        return self.__contentPanel


    @property
    def containerPanel(self):
        """Returns the ``wx.Panel`` which contains the
        :class:`.ColourBarPanel` if it is being displayed, and the content
        panel. See the note on
        :ref:`adding content <canvaspanel-adding-content>`.
        """
        return self.__containerPanel


    @property
    def colourBarCanvas(self):
        """If a colour bar is being displayed, this method returns
        the :class:`.ColourBarCanvas` instance which is used by the
        :class:`.ColourBarPanel` to render the colour bar.

        Otherwise, ``None`` is returned.
        """
        if self.__colourBar is not None:
            return self.__colourBar.getCanvas()
        return None


    def getGLCanvases(self):
        """This method must be overridden by subclasses, and must return a
        list containing all :class:`.SliceCanvas` instances which are being
        displayed.
        """
        raise NotImplementedError(
            'getGLCanvases has not been implemented '
            'by {}'.format(type(self).__name__))


    def centrePanelLayout(self):
        """Lays out the centre panel. This method may be overridden by
        sub-classes which need more advanced layout logic. See the note on
        :ref:`adding content <canvaspanel-adding-content>`
        """

        self.layoutContainerPanel()

        sizer = self.centrePanel.GetSizer()
        if sizer is not None:
            sizer.Clear()

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.__containerPanel, flag=wx.EXPAND, proportion=1)
        self.centrePanel.SetSizer(sizer)

        self.PostSizeEvent()


    def layoutContainerPanel(self):
        """Creates a ``wx.Sizer``, and uses it to lay out the colour bar panel
        and canvas panel. The sizer object is returned.

        This method is used by the default :meth:`centrePanelLayout` method,
        and is available for custom sub-class implementations to use.
        """

        sizer = self.__containerPanel.GetSizer()
        if sizer is not None:
            sizer.Clear()
            sizer = None

        sopts = self.sceneOpts

        if not sopts.showColourBar:

            if self.__colourBar is not None:
                sopts.unbindProps('colourBarLabelSide',
                                  self.__colourBar.colourBar,
                                  'labelSide')
                sopts.unbindProps('colourBarSize',
                                  self.__colourBar.canvas,
                                  'barSize')
                sopts.unbindProps('highDpi', self.__colourBar.canvas)

                self.__colourBar.destroy()
                self.__colourBar.Destroy()
                self.__colourBar = None

            sizer = wx.BoxSizer(wx.HORIZONTAL)
            sizer.Add(self.__contentPanel, flag=wx.EXPAND, proportion=1)
            self.__containerPanel.SetSizer(sizer)
            return

        if self.__colourBar is None:
            self.__colourBar = colourbarpanel.ColourBarPanel(
                self.__containerPanel,
                self.overlayList,
                self.displayCtx,
                self.frame)

            bg = sopts.bgColour
            fg = sopts.fgColour
            fs = sopts.labelSize
            self.__colourBar.colourBar.textColour = fg
            self.__colourBar.colourBar.bgColour   = bg
            self.__colourBar.colourBar.fontSize   = fs

            sopts.bindProps('colourBarLabelSide',
                            self.__colourBar.colourBar,
                            'labelSide')
            sopts.bindProps('colourBarSize',
                            self.__colourBar.canvas,
                            'barSize')
            sopts.bindProps('highDpi', self.__colourBar.canvas)

        if   sopts.colourBarLocation in ('top', 'bottom'):
            self.__colourBar.colourBar.orientation = 'horizontal'
        elif sopts.colourBarLocation in ('left', 'right'):
            self.__colourBar.colourBar.orientation = 'vertical'

        if sopts.colourBarLocation in ('top', 'bottom'):
            sizer = wx.BoxSizer(wx.VERTICAL)
        else:
            sizer = wx.BoxSizer(wx.HORIZONTAL)

        if sopts.colourBarLocation in ('top', 'left'):
            sizer.Add(self.__colourBar,    flag=wx.EXPAND)
            sizer.Add(self.__contentPanel, flag=wx.EXPAND, proportion=1)
        else:
            sizer.Add(self.__contentPanel, flag=wx.EXPAND, proportion=1)
            sizer.Add(self.__colourBar,    flag=wx.EXPAND)

        self.__containerPanel.SetSizer(sizer)


    def __colourBarPropsChanged(self, *a):
        """Called when any colour bar display properties are changed (see
        :class:`.SceneOpts`). Calls :meth:`canvasPanelLayout`.
        """
        self.centrePanelLayout()


    def __labelSizeChanged(self, *a, **kwa):
        """Called when the :class:`.SceneOpts.lablSize` changes.  If a colour
        bar is being displayed, it is updated, and the panel layout is
        refreshed.
        """
        sopts = self.sceneOpts
        if self.__colourBar is not None:
            self.__colourBar.canvas.colourBar.fontSize = sopts.labelSize
        wx.CallAfter(self.Layout)


    def __bgfgColourChanged(self, *a, **kwa):
        """Called when the :class:`.SceneOpts.bgColour` or
        :class:`.SceneOpts.fgColour` properties change.  Updates
        background/foreground colours.

        The :attr:`.SliceCanvasOpts.bgColour` properties are bound to
        ``SceneOpts.bgColour``,(see :meth:`.HasProperties.bindProps`), so we
        don't need to manually update them.

        :arg refresh: Must be passed as a keyword argument. If ``True`` (the
                      default), this ``OrthoPanel`` is refreshed.
        """

        if self.destroyed:
            return

        refresh = kwa.pop('refresh', True)

        sceneOpts = self.sceneOpts
        cpanel    = self.contentPanel
        canvases  = self.getGLCanvases()
        bg        = sceneOpts.bgColour
        fg        = sceneOpts.fgColour

        cpanel.SetBackgroundColour([round(c * 255) for c in bg])
        cpanel.SetForegroundColour([round(c * 255) for c in fg])

        if self.__colourBar is not None:
            canvas = self.__colourBar.canvas
            cbar   = self.__colourBar.colourBar
            cbar.textColour = fg
            cbar.bgColour   = bg
            canvases.append(canvas)

        if refresh:
            self.Refresh()
            self.Update()


    def __movieModeChanged(self, *a):
        """Called when the :attr:`movieMode` property changes. If it has been
        enabled, calls :meth:`__movieUpdate`, to start the movie loop.
        """

        # The fsl.utils.idle idle loop timeout
        # defaults to 200 milliseconds, which can
        # cause delays in frame updates. So when
        # movie mode is on, we bump up the rate.
        def startMovie():
            idle.idleLoop.callRate = 10
            if not self.__movieLoop(startLoop=True):
                idle.idleLoop.callRate = None

        # The __movieModeChanged method is called
        # on the props event queue. Here we make
        # sure that __movieLoop() is called *off*
        # the props event queue, by calling it from
        # the idle loop.
        if self.movieMode: idle.idle(startMovie)
        else:              idle.idleLoop.callRate = None


    def __movieLoop(self, startLoop=False):
        """Manages the triggering of the next movie frame. This method is
        called by :meth:`__movieModeChanged` when :attr:`movieMode` changes
        and when the selected overlay changes, and also by
        :meth:`__syncMovieRefresh` and :meth:`__unsyncMovieRefresh` while
        the movie loop is running, to trigger the next frame.

        :arg startLoop: This is set to ``True`` when called from
                        :meth:`__movieModeChanged`. If ``True``, and the movie
                        loop is already running, this method does nothing.

        """

        # Movie loop is already running, nothing to do.
        if startLoop and self.__movieRunning:
            return True

        # Attempt to show the next frame -
        # __movieFrame returns True if the
        # movie is continuing, False if it
        # has ended.
        self.__movieRunning = self.__movieFrame()

        return self.__movieRunning


    def canRunMovie(self, overlay, opts):
        """Returns ``True`` or ``False``, depending on whether movie mode
        is possible with the given z`overlay`` and ``opts``.
        """

        import fsl.data.image as fslimage
        import fsl.data.mesh  as fslmesh

        axis = self.movieAxis

        # 3D movies are good for all overlays
        if axis < 3:
            return True

        # 4D Nifti images are all good
        if isinstance(overlay, fslimage.Nifti) and \
           len(overlay.shape) > 3              and \
           overlay.shape[3] > 1                and \
           isinstance(opts, displayctx.VolumeOpts):
            return True

        # Mesh surfaces with N-D
        # vertex data are all good
        if isinstance(overlay, fslmesh.Mesh) and \
           opts.vertexDataLen() > 1:
            return True

        return False


    def getMovieFrame(self, overlay, opts):
        """Returns the current movie frame for the given overlay.

        A movie frame is typically a sequentially increasing number in
        some minimum/maximum range, e.g. a voxel or volume index.

        This method may be overridden by sub-classes for custom behaviour
        (e.g. the :class:`.Scene3DPanel`).
        """

        axis = self.movieAxis

        def nifti():
            if axis < 3: return opts.getVoxel(vround=False)[axis]
            else:        return opts.volume

        def mesh():
            if axis < 3: return other()
            else:        return opts.vertexDataIndex

        def other():
            return self.displayCtx.location.getPos(axis)

        import fsl.data.image as fslimage
        import fsl.data.mesh  as fslmesh

        if   isinstance(overlay, fslimage.Nifti): return nifti()
        elif isinstance(overlay, fslmesh.Mesh):   return mesh()
        else:                                     return other()


    def doMovieUpdate(self, overlay, opts):
        """Called by :meth:`__movieFrame`. Updates the properties on the
        given ``opts`` instance to move forward one frame in the movie.

        This method may be overridden by sub-classes for custom behaviour
        (e.g. the :class:`.Scene3DPanel`).

        :returns:   A value which identifies the current movie frame. This may
                    be a volume or voxel index, or a world coordinate location
                    on one axis.
        """

        axis = self.movieAxis

        def nifti():

            limit = overlay.shape[axis]

            # This method has been called off the props
            # event queue (see __movieModeChanged).
            # Therefore, all listeners on the opts.volume
            # or DisplayContext.location  properties
            # should be called immediately, in these
            # assignments.
            #
            # When the movie axis == 3 (time), this means
            # that image texture refreshes should be
            # triggered and, after the opts.volume
            # assignment, all affected GLObjects should
            # return ready() == False.
            if axis == 3:
                if opts.volume >= limit - 1: opts.volume  = 0
                else:                        opts.volume += 1

                frame = opts.volume

            else:
                voxel = opts.getVoxel()
                if voxel[axis] >= limit - 1: voxel[axis]  = 0
                else:                        voxel[axis] += 1

                self.displayCtx.location = opts.transformCoords(
                    voxel, 'voxel', 'display')

                frame = voxel[axis]
            return frame

        def mesh():

            if axis == 3:
                limit = opts.vertexDataLen()
                val   = opts.vertexDataIndex

                if val >= limit - 1: val  = 0
                else:                val += 1

                opts.vertexDataIndex = val

                return val

            else:
                return other()

        def other():

            bmin, bmax = opts.bounds.getRange(axis)
            delta      = (bmax - bmin) / 75.0

            pos = self.displayCtx.location.getPos(axis)

            if pos >= bmax: pos = bmin
            else:           pos = pos + delta

            self.displayCtx.location.setPos(axis, pos)
            return pos

        import fsl.data.image as fslimage
        import fsl.data.mesh  as fslmesh

        if   isinstance(overlay, fslimage.Nifti): frame = nifti()
        elif isinstance(overlay, fslmesh.Mesh):   frame = mesh()
        else:                                     frame = other()

        return frame


    def __movieFrame(self):
        """Called by :meth:`__movieLoop`.

        If the currently selected overlay (see
        :attr:`.DisplayContext.selectedOverlay`) is a 4D :class:`.Image` being
        displayed as a ``volume`` (see the :class:`.VolumeOpts` class), the
        :attr:`.NiftiOpts.volume` property is incremented and all
        GL canvases in this ``CanvasPanel`` are refreshed.

        :returns: ``True`` if the movie loop was started, ``False`` otherwise.
        """

        from . import scene3dpanel

        if self.destroyed:     return False
        if not self.movieMode: return False

        overlay  = self.displayCtx.getSelectedOverlay()
        canvases = self.getGLCanvases()

        if overlay is None:
            return False

        opts = self.displayCtx.getOpts(overlay)

        if not self.canRunMovie(overlay, opts):
            return False

        # We want the canvas refreshes to be
        # synchronised. So we 'freeze' them
        # while changing the image volume, and
        # then refresh them all afterwards.
        for c in canvases:
            c.FreezeDraw()
            c.FreezeSwapBuffers()

        self.doMovieUpdate(overlay, opts)

        # Now we get refs to *all* GLObjects managed
        # by every canvas - we have to wait until
        # they are all ready to be drawn before we
        # can refresh the canvases.  Note that this
        # is only necessary when the movie axis == 3
        globjs = [c.getGLObject(o)
                  for c in canvases
                  for o in self.overlayList]
        globjs = [g for g in globjs if g is not None]

        def allReady():
            return all([g.ready() for g in globjs])

        # Figure out the movie rate - the
        # number of seconds to wait until
        # triggering the next frame.
        rate    = self.movieRate
        rateMin = self.getAttribute('movieRate', 'minval')
        rateMax = self.getAttribute('movieRate', 'maxval')

        # Special case/hack - if this is a Scene3DPanel,
        # and the movie axis is X/Y/Z, we always
        # use a fast rate. Instead, the Scene3dPanel
        # will increase/decrease the rotation angle
        # to speed up/slow down the movie instead.
        if isinstance(self, scene3dpanel.Scene3DPanel) and self.movieAxis < 3:
            rate = rateMax

        rate = (rateMin + (rateMax - rate)) / 1000.0

        def update():
            self.movieSync()
            idle.idle(self.__movieLoop, after=rate)

        # Refresh the canvases when all
        # GLObjects are ready to be drawn.
        idle.idleWhen(update, allReady, pollTime=rate / 10)

        return True


    def movieSync(self):
        """Called by :meth:`__movieUpdate`. Updates all GL canvases, attempting
        to refresh them in a synchronised manner.

        Ideally all canvases should be drawn off-screen (i.e. rendered to the
        back buffer), and then all refreshed together (back and front buffers
        swapped). Unfortunately some OpenGL drivers seem to have trouble with
        this approach, and require drawing and front/back buffer swaps to be
        done at the same time.

        This method will refresh the GL canvases in either a synchronised or
        unsynchronised manner, depending upon the value of the
        :attr:`movieSyncRefresh` property.
        """

        canvases = self.getGLCanvases()

        if self.movieSyncRefresh:
            for c in canvases:
                c.ThawDraw()
                c.Refresh()

            for c in canvases:
                c.ThawSwapBuffers()
                c.SwapBuffers()
        else:
            for c in canvases:
                c.ThawDraw()
                c.ThawSwapBuffers()
                c.Refresh()



#!/usr/bin/env python
#
# loadoverlayfromdir.py - Action which allows the user to load overlays
#                         specified with a directory.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadOverlayFromDirAction`, which allows
the user to load overlay files, specified with a directory, into the
:class:`.OverlayList`.
"""


import fsleyes.autodisplay as autodisplay
from . import                 loadoverlay
from . import                 base


class LoadOverlayFromDirAction(base.Action):
    """The ``LoadOverlayFromDirAction`` allows the user to add overlays
    to the :class:`.OverlayList`. This functionality is provided by functions
    in the :mod:`.loadoverlay` module.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create an ``OpenDirAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__openDir)


    def __openDir(self):
        """Calls the :func:`.loadoverlay.interactiveLoadOverlays` method.

        If overlays were added, updates the
        :attr:`.DisplayContext.selectedOverlay` accordingly.

        If :attr:`.DisplayContext.autoDisplay` is ``True``, uses the
        :mod:`.autodisplay` module to configure the display properties
        of each new overlay.
        """

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)
            self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[-1]

            if self.displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.overlayList,
                                            self.displayCtx)

        loadoverlay.interactiveLoadOverlays(
            dirdlg=True,
            onLoad=onLoad,
            inmem=self.displayCtx.loadInMemory)



#!/usr/bin/env python
#
# wxglslicecanvas.py - The WXGLSliceCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WXGLSliceCanvas` class, which is a
:class:`.SliceCanvas` for use in a :mod:`wx` application.
"""


import six

import wx
import wx.glcanvas as wxgl

import fsleyes_widgets as     fwidgets
import fsleyes_props   as     props

import fsleyes.gl      as     fslgl
from   .               import slicecanvas


class WXGLSliceCanvas(six.with_metaclass(fslgl.WXGLMetaClass,
                                         slicecanvas.SliceCanvas,
                                         fslgl.WXGLCanvasTarget,
                                         wxgl.GLCanvas)):
    """The ``WXGLSliceCanvas`` is a :class:`.SliceCanvas`, a
    :class:`wx.glcanvas.GLCanvas` and a :class:`.WXGLCanvasTarget`. If you
    want to use a :class:`.SliceCanvas` in your :mod:`wx` application, then
    you should use a ``WXGLSliceCanvas``.

    .. note:: The ``WXGLSliceCanvas`` assumes the existence of the
              :meth:`.SliceCanvas._updateDisplayBounds` method.
    """

    def __init__(self, parent, overlayList, displayCtx, zax=0):
        """Create a ``WXGLSliceCanvas``. See :meth:`.SliceCanvas.__init__` for
        details on the arguments.
        """

        attrs = fslgl.WXGLCanvasTarget.displayAttribues()

        wxgl.GLCanvas          .__init__(self, parent, **attrs)
        fslgl.WXGLCanvasTarget .__init__(self)
        slicecanvas.SliceCanvas.__init__(self, overlayList, displayCtx, zax)

        self.Bind(wx.EVT_SIZE, self.__onResize)


    def destroy(self):
        """Must be called when this ``WXGLSliceCanvas`` is no longer needed.
        Clears some event listeners and calls the base class ``destroy``
        method.
        """
        self.Unbind(wx.EVT_SIZE)
        slicecanvas.SliceCanvas.destroy(self)
        fslgl.WXGLCanvasTarget .destroy(self)


    def __onResize(self, ev):
        """Called on ``wx.EVT_SIZE`` events, when the canvas is resized. When
        the canvas is resized, we have to update the display bounds to preserve
        the aspect ratio.
        """
        ev.Skip()

        with props.skip(self.opts, 'displayBounds', self.name):
            centre = self.getDisplayCentre()
            self._updateDisplayBounds()
            self.centreDisplayAt(*centre)


    def Show(self, show):
        """Overrides ``GLCanvas.Show``. When running over SSH/X11, it doesn't
        seem to be possible to hide a ``GLCanvas`` - the most recent scene
        displayed on the canvas seems to persist, does not get overridden, and
        gets drawn on top of other things in the interface:

        .. image:: images/x11_slicecanvas_show_bug.png
           :scale: 50%
           :align: center

        This is not ideal, and I have no idea why it occurs. The only
        workaround that I've found to work is, instead of hiding the canvas,
        to set its size to 0. So this method does just that.
        """

        # If not in SSH, we can just
        # show/hide normally.
        if not fwidgets.inSSHSession():
            wxgl.GLCanvas.Show(self, show)

        elif not show:
            self.SetMinSize((0, 0))
            self.SetMaxSize((0, 0))
            self.SetSize(   (0, 0))


    def Hide(self):
        """Overrides ``GLCanvas.Hide``. Calls :meth:`Show`. """
        self.Show(False)



#!/usr/bin/env python
#
# scene3dpanel.py - The Scene3DPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Scene3DPanel` class, a FSLeyes view which
draws the scene in 3D.
"""


import logging

import wx

import numpy as np

import fsl.transform.affine                as affine
import fsleyes.displaycontext.scene3dopts  as scene3dopts
import fsleyes.gl.wxglscene3dcanvas        as scene3dcanvas
import fsleyes.profiles.scene3dviewprofile as scene3dviewprofile
import fsleyes.actions                     as actions
from . import                                 canvaspanel


log = logging.getLogger(__name__)


class Scene3DPanel(canvaspanel.CanvasPanel):
    """The ``Scene3DPanel`` is a :class:`.CanvasPanel` which draws the
    contents of the :class:`.OverlayList` as a 3D scene.


    The ``Scene3DPanel`` uses a :class:`.Scene3DCanvas`, which manages all of
    the GL state and drawing logic. A :class:`.Scene3DViewProfile` instance
    is used to manage all of the user interaction logic.


    The scene properties are described and changed via a :class:`.Scene3DOpts`
    instance, accessible through the :meth:`.CanvasPanel.sceneOpts`
    property.
    """


    @staticmethod
    def defaultLayout():
        """Returns a list of control panel types to be added for the default
        3D panel layout.
        """
        return ['OverlayDisplayToolBar',
                'Scene3DToolBar',
                'OverlayListPanel',
                'LocationPanel']


    @staticmethod
    def controlOrder():
        """Returns a list of control panel names, specifying the order in
        which they should appear in the  FSLeyes ortho panel settings menu.
        """
        return ['OverlayListPanel',
                'LocationPanel',
                'OverlayInfoPanel',
                'OverlayDisplayPanel',
                'CanvasSettingsPanel',
                'AtlasPanel',
                'OverlayDisplayToolBar',
                'Scene3DToolBar',
                'FileTreePanel']


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``Scene3dPanel``.

        :arg parent:      A :mod:`wx` parent object.
        :arg overlayList: A :class:`.OverlayList` instance.
        :arg displayCtx:  A :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        sceneOpts = scene3dopts.Scene3DOpts(self)

        canvaspanel.CanvasPanel.__init__(self,
                                         parent,
                                         overlayList,
                                         displayCtx,
                                         frame,
                                         sceneOpts)


        # In 3D, the displaySpace must always be
        # set to world, regardless of the parent
        # DC value. This can be overridden manually
        # however (e.g. through the python shell)
        displayCtx.detachDisplaySpace()
        displayCtx.defaultDisplaySpace = 'world'
        displayCtx.displaySpace        = 'world'

        contentPanel = self.contentPanel

        self.__canvas = scene3dcanvas.WXGLScene3DCanvas(contentPanel,
                                                        overlayList,
                                                        displayCtx)

        opts = self.__canvas.opts

        opts.bindProps('pos',           displayCtx, 'location')
        opts.bindProps('showCursor',    sceneOpts)
        opts.bindProps('cursorColour',  sceneOpts)
        opts.bindProps('bgColour',      sceneOpts)
        opts.bindProps('showLegend',    sceneOpts)
        opts.bindProps('legendColour',  sceneOpts, 'fgColour')
        opts.bindProps('labelSize',     sceneOpts)
        opts.bindProps('occlusion',     sceneOpts)
        opts.bindProps('light',         sceneOpts)
        opts.bindProps('lightPos',      sceneOpts)
        opts.bindProps('lightDistance', sceneOpts)
        opts.bindProps('showLight',     sceneOpts)
        opts.bindProps('zoom',          sceneOpts)
        opts.bindProps('offset',        sceneOpts)
        opts.bindProps('rotation',      sceneOpts)
        opts.bindProps('highDpi',       sceneOpts)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.__canvas, flag=wx.EXPAND, proportion=1)
        contentPanel.SetSizer(sizer)

        self.centrePanelLayout()
        self.initProfile(scene3dviewprofile.Scene3DViewProfile)
        self.syncLocation = True


    def destroy(self):
        """Must be called when this ``Scene3DPanel`` is no longer in use.
        """

        self.__canvas.destroy()
        self.__canvas = None
        canvaspanel.CanvasPanel.destroy(self)


    def getGLCanvases(self):
        """Returns all of the :class:`.SliceCanvas` instances contained
        within this ``Scene3DPanel``.
        """
        return [self.__canvas]


    def getActions(self):
        """Overrides :meth:`.ViewPanel.getActions`. Returns a list of actions
        that can be executed on this ``Scene3DPanel``, and which will be added
        to its view menu.
        """
        actionz = [self.screenshot,
                   self.movieGif,
                   self.showCommandLineArgs,
                   self.applyCommandLineArgs,
                   None,
                   self.toggleDisplaySync,
                   self.resetDisplay]

        names = [a.actionName if a is not None else None for a in actionz]
        return list(zip(names, actionz))


    @actions.action
    def resetDisplay(self):
        """An action which resets the current camera configuration
        (zoom/pan/rotation). See the :meth:`.Scene3DViewProfile.resetDisplay`
        method.
        """
        self.currentProfile.resetDisplay()


    def getMovieFrame(self, overlay, opts):
        """Returns the current movie frame. If the :attr:`movieAxis` is ``3``
        (e.g. time series), the volume index is returned. Otherwise the
        current rotation matrix is returned.
        """
        if self.movieAxis == 3:
            return super(Scene3DPanel, self).getMovieFrame(overlay, opts)
        else:
            return np.copy(self.__canvas.opts.rotation)


    def doMovieUpdate(self, overlay, opts):
        """Overrides :meth:`.CanvasPanel.doMovieUpdate`. For x/y/z axis
        movies, the scene is rotated. Otherwise (for time) the ``CanvasPanel``
        implementation is called.
        """

        if self.movieAxis >= 3:
            return canvaspanel.CanvasPanel.doMovieUpdate(self, overlay, opts)
        else:

            canvas  = self.__canvas
            currot  = canvas.opts.rotation
            rate    = float(self.movieRate)
            rateMin = self.getAttribute('movieRate', 'minval')
            rateMax = self.getAttribute('movieRate', 'maxval')
            rate    = 0.1 + 0.9 * (rate - rateMin) / (rateMax - rateMin)
            rate    = rate * np.pi / 10

            rots                 = [0, 0, 0]
            rots[self.movieAxis] = rate

            xform = affine.axisAnglesToRotMat(*rots)
            xform = affine.concat(xform, currot)

            canvas.opts.rotation = xform
            return np.copy(xform)



#!/usr/bin/env python
#
# removeoverlay.py - Action which removes the current overlay.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`RemoveOverlayAction` class, and the
:func:`removeOverlay` function, which provides logic to remove an overlay
from the :class:`.OverlayList`.
"""


import fsl.data.image  as fslimage
import fsleyes.strings as strings
from . import             base


class RemoveOverlayAction(base.Action):
    """The ``RemoveOverlayAction`` allows the uesr to remove the currently
    selected overlay.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``RemoveOverlayAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__removeOverlay)

        self.__name = '{}_{}'.format(type(self).__name__, id(self))

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)


    def destroy(self):
        """Must be called when this ``RemoveOverlayAction`` is no longer
        needed. Removes property listeners, and then calls
        :meth:`.Action.destroy`.
        """
        self.overlayList.removeListener('overlays', self.__name)
        base.Action.destroy(self)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Updates the
        :attr:`.Action.enabled` flag
        """
        self.enabled = len(self.overlayList) > 0


    def __removeOverlay(self):
        """Removes the currently selected overlay (as defined by the
        :attr:`.DisplayContext.selectedOverlay) from the :class:`.OverlayList`.
        """
        removeOverlay(self.overlayList, self.displayCtx)


def removeOverlay(overlayList, displayCtx, overlay=None, stringKey=None):
    """Removes the specified overlay (or the currently selected overlay,
    if ``overlay is None``) from the overlay list. If the overlay is not
    saved, the user is prompted to confirm the removal.

    :arg overlay:   Overlay to remove. If ``None``, the currently selected
                    overlay is removed.

    :arg stringKey: Key to use in the :mod:`.strings` module for the
                    dialog with which the user is prompted if the overlay
                    has unsaved changes.

    :returns:       ``True`` if the overlay was removed, ``False`` otherise.
    """

    import wx

    if overlay is None:
        overlay = displayCtx.getSelectedOverlay()

    if stringKey is None:
        stringKey = 'removeoverlay.unsaved'

    if isinstance(overlay, fslimage.Image) and not overlay.saveState:

        msg    = strings.messages[stringKey]
        title  = strings.titles[  stringKey]
        parent = wx.GetApp().GetTopWindow()

        dlg = wx.MessageDialog(parent,
                               message=msg,
                               caption=title,
                               style=(wx.YES_NO        |
                                      wx.NO_DEFAULT    |
                                      wx.CENTRE        |
                                      wx.ICON_WARNING))

        dlg.CentreOnParent()
        if dlg.ShowModal() == wx.ID_NO:
            return False

    overlayList.remove(overlay)
    return True



#!/usr/bin/env python
#
# imagetexture.py - The ImageTexture and ImageTexture2D class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ImageTexture` and :class:`.ImageTexture2D`
classes, :class:`.Texture3D` and :class:`.Texture2D` classes for storing an
:class:`.Image` instance.
"""


import logging
import contextlib
import collections.abc as abc

import numpy as np

import fsl.transform.affine  as affine
import fsl.data.imagewrapper as imagewrapper
import fsleyes_widgets       as fwidgets
from . import data           as texdata
from . import                   texture2d
from . import                   texture3d


log = logging.getLogger(__name__)


def createImageTexture(name, image, *args, **kwargs):
    """Creates and returns an appropriate texture type (either
    :class:`ImageTexture` or :class:`ImageTexture2D`) for the given image.
    """

    ndims = texdata.numTextureDims(image.shape[:3])

    if ndims == 3: return ImageTexture(  name, image, *args, **kwargs)
    else:          return ImageTexture2D(name, image, *args, **kwargs)


class ImageTextureBase(object):
    """Base class shared by the :class:`ImageTexture` and
    :class:`ImageTexture2D` classes. Contains logic for retrieving a
    specific volume from a 3D + time or 2D + time :class:`.Image`, and
    for retrieving a specific channel from an RGB(A) ``Image``.
    """


    @staticmethod
    def validateShape(image, texnvals, texndims):
        """Called by :meth:`__init__`. Makes sure that the specified texture
        settings (number of dimensions, and number of values per texture
        element) are compatible with the image.

        :arg image:    :class:`.Image`
        :arg texnvals: Number of values per texture element
        :arg texndims: Number of texture dimensions
        :raises:       :exc:`RuntimeError` if the texture properties are not
                       compatible with the image
        """

        imgnvals = image.nvals
        imgndims = len(image.shape)

        # Anything goes for single-
        # valued textures
        if texnvals == 1:
            return

        # For multi-valued 4D textures,
        # the image must have a shape
        # of the form:
        # (x, y, z, [1, [1, [1, [1, ]]]] nvals)
        if imgnvals == 1:
            expShape     = list(image.shape[:3])
            expShape    += [1] * (imgndims - 3)
            expShape[-1] = texnvals
            if list(image.shape) != expShape:
                raise RuntimeError(
                    'Data shape mismatch: texture size {} requested for '
                    'image shape {}'.format(texnvals, image.shape))

        # Or must be a structured, i.e.
        # RGB(A) array with the correct
        # number of values per voxel.
        elif imgnvals != texnvals:
            raise RuntimeError(
                'Data shape mismatch: texture size {} requested for '
                'image with nvals {}'.format(texnvals, imgnvals))


    def __init__(self, image, nvals, ndims):
        """Create an ``ImageTextureBase``

        :arg image: The :class:`.Image`
        :arg nvals: Number of values per texture element
        :arg ndims: Number of texture dimensions
        """

        self.validateShape(image, nvals, ndims)

        self.__name    = 'ImageTextureBase_{}'.format(id(self))
        self.__image   = image
        self.__volume  = None
        self.__channel = None

        self.__image.register(self.__name,
                              self.__imageDataChanged,
                              'data',
                              runOnIdle=True)


    def destroy(self):
        """Must be called when this ``ImageTextureBase`` is no longer needed.
        """
        self.__image.deregister(self.__name, 'data')
        self.__image = None


    @property
    def image(self):
        """Returns the :class:`.Image` managed by this ``ImageTextureBase``.
        """
        return self.__image


    @property
    def volume(self):
        """For :class:`.Image` instances with more than three dimensions,
        specifies the indices for the fourth and above dimensions with which
        to extract the 3D texture data. If the image has four dimensions, this
        may be a scalar, otherwise it must be a sequence of
        (``Image.ndim - 3``) the correct length.
        """
        return self.__volume


    @volume.setter
    def volume(self, volume):
        """Set the current volume. """
        self.set(volume=volume)


    @property
    def channel(self):
        """For :class:`.Image` instances with multiple values per voxel, such
        as ``RGB24`` or ``RGBA32`` images, this option allows the channel
        to be selected.
        """
        return self.__channel


    @channel.setter
    def channel(self, channel):
        """Set the current channel. """
        self.set(channel=channel)


    def prepareSetArgs(self, **kwargs):
        """Called by sub-classes in their :meth:`.Texture2D.set`/
        :meth:`.Texture3D.set` override.

        Prepares arguments to be passed through to the underlying
        ``set`` method.  This method accepts any parameters that are accepted
        by :meth:`.Texture3D.set`, plus the following:

        =============== ======================================================
        ``volume``      See :meth:`volume`.

        ``channel``     See :meth:`channel`.

        ``volRefresh``  If ``True`` (the default), the texture data will be
                        refreshed even if the ``volume`` and ``channel``
                        parameters haven't changed. Otherwise, if ``volume``
                        and ``channel`` haven't changed, the texture will not
                        be refreshed.
        =============== ======================================================

        :returns: ``True`` if any settings have changed and the
                  ``ImageTexture`` is to be refreshed , ``False`` otherwise.
        """

        kwargs             .pop('data',           None)
        normRange  = kwargs.pop('normaliseRange', None)
        channel    = kwargs.pop('channel',        self.__channel)
        volume     = kwargs.pop('volume',         self.__volume)
        volRefresh = kwargs.pop('volRefresh',     True)
        image      = self.image
        nvals      = self.nvals
        ndims      = image.ndim

        if len(image.shape) == 3: volume  = None
        if image.nvals      == 1: channel = None

        if normRange is None:
            normRange = image.dataRange

        if ndims == 3 or nvals > 1:
            volume = None
        else:
            if volume is None and self.__volume is None:
                volume = [0] * (ndims - 3)

            elif not isinstance(volume, abc.Sequence):
                volume = [volume]

            if len(volume) != ndims - 3:
                raise ValueError('Invalid volume indices for {} '
                                 'dims: {}'.format(ndims, volume))

        if (not volRefresh)           and \
           (volume  == self.__volume) and \
           (channel == self.__channel):
            return kwargs

        self.__volume  = volume
        self.__channel = channel

        data = self.__getData(volume, channel)
        data = self.shapeData(data)

        kwargs['data']           = data
        kwargs['normaliseRange'] = normRange

        return kwargs


    def __getData(self, volume, channel):
        """Extracts data from the :class:`.Image` for use as texture data.

        For textures with multiple values per element (either by volume, or by
        channel), the data is arranged appropriately, i.e. with the value as
        the first dimension.

        :arg volume:  Volume index/indices, for images with more than three
                      dimensions.
        :arg channel: Channel, for RGB(A) images.
        """

        image = self.image
        slc   = [slice(None), slice(None), slice(None)]

        if volume is not None:
            slc += volume

        if channel is None: data = self.image[              tuple(slc)]
        else:               data = self.image.data[channel][tuple(slc)]

        # For single-valued textures,
        # we don't need to do anything
        if self.nvals == 1:
            return data

        # Multi-valued texture with a
        # single-valued 4D image - we
        # assume that each volume
        # corresponds to a channel
        elif image.nvals == 1 and volume is None:
            data = data.transpose((3, 0, 1, 2))

        # Multi-valued RGB(A) texture
        # with a multi-valued image -
        # cast/reshape the image data
        elif image.nvals > 1 and channel is None:
            data = np.ndarray(buffer=data.data,
                              shape=[self.nvals] + list(data.shape),
                              order='F',
                              dtype=np.uint8)

        return data


    def __imageDataChanged(self, image, topic, sliceobj):
        """Called when the :class:`.Image` notifies about a data changes.
        Triggers an image texture refresh via a call to :meth:`set`.

        :arg image:    The ``Image`` instance

        :arg topic:    The string ``'data'``

        :arg sliceobj: Slice object specifying the portion of the image
                       that was changed.
        """

        # TODO If the change has caused the image
        #      data range to change, and texture
        #      data normalisation is on, you have
        #      to refresh the full texture.
        #
        #      The Image instance does follow up
        #      data change notifications with a
        #      data range notification; perhaps
        #      you can use this somehow.

        # If the data change was performed using
        # normal array indexing, we can just replace
        # that part of the image texture.
        if isinstance(sliceobj, tuple):

            # Get the new data, and calculate an
            # offset into the full image from the
            # slice object.
            data   = np.array(image[sliceobj])
            offset = imagewrapper.sliceObjToSliceTuple(sliceobj, image.shape)
            offset = [o[0] for o in offset]

            # Make sure the data/offset are
            # compatible with 2D textures
            data       = self.shapeData(data, oldShape=image.shape)
            offset[:3] = affine.transform(
                offset[:3], self.texCoordXform(image.shape))

            log.debug('{} data changed - refreshing part of '
                      'texture (offset: {}, size: {})'.format(
                          image.name,
                          offset, data.shape))

            self.patchData(data, offset)

        # Otherwise (boolean array indexing) we have
        # to replace the whole image texture.
        else:

            log.debug('{} data changed - refreshing full '
                      'texture'.format(image.name))

            self.set()


class ImageTexture(ImageTextureBase, texture3d.Texture3D):
    """The ``ImageTexture`` class contains the logic required to create and
    manage a 3D texture which represents a :class:`.Image` instance.


    Once created, the :class:`.Image` instance is available as an attribute of
    an ``ImageTexture`` object, called ``image``. See the :class:`.Texture3D`
    documentation for more details.


    For multi-valued (e.g. RGB) textures, the :class:`.Texture3D` class
    requires data to be passed as a ``(C, X, Y, Z)`` array (for ``C`` values).
    If an ``ImageTexture`` is created with an image of type
    ``NIFTI_TYPE_RGB24`` or ``NIFTI_TYPE_RGBA32``, it will take care of
    re-arranging the image data so that it has the shape required by the
    ``Texture3D`` class.
    """


    threadedDefault = None
    """Default value used for the ``threaded`` argument passed to
    :meth:`__init__`. When this is set to ``None``, the default value will be
    the value of :func:`.fsleyes_widgets.haveGui`.
    """


    @classmethod
    @contextlib.contextmanager
    def enableThreading(cls, enable=True):
        """Context manager which can be used to temporarily set the
        default value of the ``threaded`` argument passedto :meth:`__init__`.
        """

        oldval = ImageTexture.threadedDefault
        ImageTexture.threadedDefault = enable

        try:
            yield
        finally:
            ImageTexture.threadedDefault = oldval


    def __init__(self,
                 name,
                 image,
                 **kwargs):
        """Create an ``ImageTexture``. A listener is added to the
        :attr:`.Image.data` property, so that the texture data can be
        refreshed whenever the image data changes - see the
        :meth:`__imageDataChanged` method.

        :arg name:   A name for this ``imageTexure``.

        :arg image:  The :class:`.Image` instance.

        :arg volume: Initial volume index/indices, for >3D images.

        All other arguments are passed through to the
        :meth:`.Texture3D.__init__` method, and thus used as initial texture
        settings.


        .. note:: The default value of the ``threaded`` parameter is set to
                  the value of :attr:`threadedDefault`.
        """

        nvals              = kwargs.get('nvals', 1)
        kwargs['nvals']    = nvals
        kwargs['scales']   = image.pixdim[:3]
        kwargs['border']   = [0, 0, 0, 0]
        kwargs['threaded'] = kwargs.get('threaded',
                                        ImageTexture.threadedDefault)

        if kwargs['threaded'] is None:
            kwargs['threaded'] = fwidgets.haveGui()

        ImageTextureBase   .__init__(self, image, nvals, 3)
        texture3d.Texture3D.__init__(self, name, **kwargs)


    def destroy(self):
        """Must be called when this ``ImageTexture`` is no longer needed."""
        texture3d.Texture3D.destroy(self)
        ImageTextureBase   .destroy(self)


    def set(self, **kwargs):
        """Overrides :meth:`.Texture3D.set`.  Passes all arguments through the
        :meth:`prepareSetArgs` method, then passes them on to
        :meth:`.Texture3D.set`.

        :returns: ``True`` if any settings have changed and the
                  ``ImageTexture`` is to be refreshed , ``False`` otherwise.
        """
        return texture3d.Texture3D.set(self, **self.prepareSetArgs(**kwargs))


class ImageTexture2D(ImageTextureBase, texture2d.Texture2D):
    """The ``ImageTexture2D`` class is the 2D analogue of the
    :class:`ImageTexture` class, for managing a 2D texture which represents an
    :class:`.Image` instance.
    """


    def __init__(self,
                 name,
                 image,
                 **kwargs):
        """Create an ``ImageTexture2D``. """

        nvals            = kwargs.get('nvals', 1)
        kwargs['nvals']  = nvals
        kwargs['border'] = [0, 0, 0, 0]
        kwargs['scales'] = image.pixdim[:3]

        ImageTextureBase   .__init__(self, image, nvals, 2)
        texture2d.Texture2D.__init__(self, name, **kwargs)


    def destroy(self):
        """Must be called when this ``ImageTexture2D`` is no longer needed. """
        texture2d.Texture2D.destroy(self)
        ImageTextureBase   .destroy(self)


    def set(self, **kwargs):
        """Overrides :meth:`.Texture2D.set`.  Passes all arguments through the
        :meth:`prepareSetArgs` method, then passes them on to
        :meth:`.Texture2D.set`.

        :returns: ``True`` if any settings have changed and the
                  ``ImageTexture`` is to be refreshed , ``False`` otherwise.
        """
        return texture2d.Texture2D.set(self, **self.prepareSetArgs(**kwargs))



#!/usr/bin/env python
#
# annotations.py - 2D annotations on a SliceCanvas.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Annotations` class, which implements
functionality to draw 2D OpenGL annotations on a :class:`.SliceCanvas`.


The :class:`Annotations` class is used by the :class:`.SliceCanvas` and
:class:`.LightBoxCanvas` classes, and users of those class, to annotate the
canvas.


.. note:: The ``Annotations`` class only works with the :class:`.SliceCanvas`
          and :class:`.LightBoxCanvas` - there is no support for the
          :class:`.Scene3DCanvas`.


All annotations derive from the :class:`AnnotationObject` base class. The
following annotation types are defined:

.. autosummary::
   :nosignatures:

   Point
   Line
   Arrow
   Rect
   Ellipse
   VoxelSelection
   TextAnnotation
"""


import time
import logging

import numpy       as np
import OpenGL.GL   as gl

import fsl.transform.affine     as affine
import fsleyes_props            as props
import fsleyes.gl.globject      as globject
import fsleyes.gl.routines      as glroutines
import fsleyes.gl.resources     as glresources
import fsleyes.gl.textures      as textures
import fsleyes.gl.text          as gltext
import fsleyes.gl.textures.data as texdata


log = logging.getLogger(__name__)


class Annotations(props.HasProperties):
    """An :class:`Annotations` object provides functionality to draw 2D
    annotations on a :class:`.SliceCanvas`. Annotations may be enqueued via
    any of the :meth:`line`, :meth:`rect`, :meth:`ellpse`, :meth:`point`,
    :meth:`selection` or :meth:`obj`, methods, and de-queued via the
    :meth:`dequeue` method.


    Annotations can be enqueued in one of three ways, using the `hold` and
    `fixed` parameters:


      - **Transient**: When calling ``line``, ``rect``, etc, passing
        ``hold=False` enqueues the annotation for the next call to
        :meth:`draw`. After the annotation is drawn, it is removed from the
        queue, and would need to be re-queued to draw it again. The ``fixed``
        parameter has no effect for transient annotations.

      - **Fixed**: When calling ``line``, ``rect``, etc, passing
        ``hold=True`` and ```fixed=True`` enqueues the annotation for all
        subsequent calls to :meth:`draw`. Fixed annotations are stored in
        an internal, inaccessible queue, so if you need to manipulate a
        fixed annotation, you need to maintain your own reference to it.

      - `**Persistent`**: When calling ``line``, ``rect``, etc, passing
        ``hold=True`` and ``fixed=False`` adds the annotation to the
        accessible :attr:`annotations` list.


    Transient annotations are intended for one-off annotations, e.g. a
    cursor mark at the current mouse location.


    Fixed annotations are intended for persistent annotations which are
    intended to be immutable, i.e. that cannot be directly manipulated by the
    user, e.g. anatomical orientation labels on the canvases of an
    :class:`.OrthoPanel`.


    Persistent annotations are intended for persistent annotations which are
    intended to be manipulated by the user - these annotations are used by the
    :class:`.AnnotationPanel` in conjunction with the
    :class:`.OrthoAnnotateProfile`.


    After annotations have been enqueued in one of the above manners, a call
    to :meth:`draw` will draw each annotation on the canvas, and clear the
    transient queue. The default value for ``hold`` is ``False``, and ``fixed``
    is ``True``,


    Annotations can be queued by one of the helper methods on the
    :class:`Annotations` object (e.g. :meth:`line` or :meth:`rect`), or by
    manually creating an :class:`AnnotationObject` and passing it to the
    :meth:`obj` method.
    """


    annotations = props.List()
    """Contains all persistent :class:`AnnotationObject` instances, which have
    been added to the queue with ``hold=True`` and ``fixed=False``.
    """


    def __init__(self, canvas, xax, yax):
        """Creates an :class:`Annotations` object.

        :arg canvas: The :class:`.SliceCanvas` that owns this
                     ``Annotations`` object.

        :arg xax:    Index of the display coordinate system axis that
                     corresponds to the horizontal screen axis.

        :arg yax:    Index of the display coordinate system axis that
                     corresponds to the vertical screen axis.
        """

        self.__transient = []
        self.__fixed     = []
        self.__xax       = xax
        self.__yax       = yax
        self.__zax       = 3 - xax - yax
        self.__canvas    = canvas


    @property
    def canvas(self):
        """Returns a ref to the canvas that owns this ``Annotations`` instance.
        """
        return self.__canvas


    def setAxes(self, xax, yax):
        """This method must be called if the display orientation changes.  See
        :meth:`__init__`.
        """

        self.__xax = xax
        self.__yax = yax
        self.__zax = 3 - xax - yax


    def getDisplayBounds(self):
        """Returns a tuple containing the ``(xmin, xmax, ymin, ymax)`` display
        bounds of the ``SliceCanvas`` that owns this ``Annotations`` object.
        """
        return self.__canvas.opts.displayBounds


    def line(self, *args, **kwargs):
        """Queues a line for drawing - see the :class:`Line` class. """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = Line(self, *args, **kwargs)
        return self.obj(obj, hold, fixed)


    def arrow(self, *args, **kwargs):
        """Queues an arrow for drawing - see the :class:`Arrow` class. """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = Arrow(self, *args, **kwargs)
        return self.obj(obj, hold, fixed)


    def point(self, *args, **kwargs):
        """Queues a point for drawing - see the :class:`Point` class. """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = Point(self, *args, **kwargs)
        return self.obj(obj, hold, fixed)


    def rect(self, *args, **kwargs):
        """Queues a rectangle for drawing - see the :class:`Rectangle` class.
        """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = Rect(self, *args, **kwargs)
        return self.obj(obj, hold, fixed)


    def ellipse(self, *args, **kwargs):
        """Queues a circle for drawing - see the :class:`Ellipse` class.
        """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = Ellipse(self, *args, **kwargs)

        return self.obj(obj, hold, fixed)


    def selection(self, *args, **kwargs):
        """Queues a selection for drawing - see the :class:`VoxelSelection`
        class.
        """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = VoxelSelection(self, *args, **kwargs)
        return self.obj(obj, hold, fixed)


    def text(self, *args, **kwargs):
        """Queues a text annotation for drawing - see the :class:`Text`
        class.
        """
        hold  = kwargs.pop('hold',  False)
        fixed = kwargs.pop('fixed', True)
        obj   = TextAnnotation(self, *args, **kwargs)
        return self.obj(obj, hold, fixed)


    def obj(self, obj, hold=False, fixed=True):
        """Queues the given :class:`AnnotationObject` for drawing.

        :arg hold:  If ``True``, the given ``AnnotationObject`` will be added
                    to the fixed or persistent queues, and will remain there
                    until it is explicitly removed. Otherwise (the default),
                    the object will be added to the transient queue, and
                    removed from the queue after it has been drawn.

        :arg fixed: If ``True`` (the default), and ``hold=True``, the given
                    ``AnnotationObject`` will be added to the fixed queue, and
                    will remain there until it is explicitly
                    removed. Otherwise, the object will be added to the
                    persistent queue and, again, will remain there until it is
                    explicitly removed. Has no effect when ``hold=False``.
        """
        if   hold and fixed: self.__fixed    .append(obj)
        elif hold:           self.annotations.append(obj)
        else:                self.__transient.append(obj)
        return obj


    def dequeue(self, obj, hold=False, fixed=True):
        """Removes the given :class:`AnnotationObject` from the appropriate
        queue, but does not call its :meth:`.GLObject.destroy` method - this
        is the responsibility of the caller.
        """

        if hold and fixed:
            try:               self.__fixed.remove(obj)
            except ValueError: pass
        elif hold:
            try:               self.annotations.remove(obj)
            except ValueError: pass
        else:
            try:               self.__transient.remove(obj)
            except ValueError: pass


    def clear(self):
        """Clears all queues, and calls the :meth:`.GLObject.destroy` method
        on every object in the queue.
        """

        for obj in self.__fixed:     obj.destroy()
        for obj in self.__transient: obj.destroy()
        for obj in self.annotations: obj.destroy()

        self.__fixed        = []
        self.__transient    = []
        self.annotations[:] = []


    def draw(self, zpos, xform=None):
        """Draws all enqueued annotations. Fixed annotations are drawn first,
        then persistent, then transient - i.e. transient annotations will
        be drawn on top of persistent, which will be drawn on to of fixed.

        :arg zpos:     Position along the Z axis, above which all annotations
                       should be drawn.

        :arg xform:    Transformation matrix which should be applied to all
                       objects.
        """

        objs = (list(self.__fixed)     +
                list(self.annotations) +
                list(self.__transient))

        if xform is not None:
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPushMatrix()
            gl.glMultMatrixf(xform.ravel('F'))

        drawTime = time.time()
        axes     = (self.__xax, self.__yax, self.__zax)

        for obj in objs:

            if obj.expired(drawTime): continue
            if not obj.enabled:       continue
            if obj.honourZLimits:
                if obj.zmin is not None and zpos < obj.zmin: continue
                if obj.zmax is not None and zpos > obj.zmax: continue

            if obj.xform is not None:
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPushMatrix()
                gl.glMultMatrixf(obj.xform.ravel('F'))

            if obj.colour is not None:

                if len(obj.colour) == 3: colour = list(obj.colour) + [1.0]
                else:                    colour = list(obj.colour)

                colour[3] = obj.alpha / 100.0

                gl.glColor4f(*colour)

            if obj.lineWidth is not None:
                gl.glLineWidth(obj.lineWidth)

            try:
                obj.preDraw()
                obj.draw2D(zpos, axes)
                obj.postDraw()
            except Exception as e:
                log.warning('{}'.format(e), exc_info=True)

            if obj.xform is not None:
                gl.glPopMatrix()

        if xform is not None:
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPopMatrix()

        # Clear the transient queue after each draw
        self.__transient = []


class AnnotationObject(globject.GLSimpleObject, props.HasProperties):
    """Base class for all annotation objects. An ``AnnotationObject`` is drawn
    by an :class:`Annotations` instance. The ``AnnotationObject`` contains some
    attributes which are common to all annotation types:

    ============= ============================================================
    ``colour``    Annotation colour
    ``alpha``     Transparency
    ``enabled``   Whether the annotation should be drawn or not.
    ``lineWidth`` Annotation line width (if the annotation is made up of
                  lines)
    ``xform``     Custom transformation matrix to apply to annotation
                  vertices.
    ``expiry``    Time (in seconds) after which the annotation will expire and
                  not be drawn.
    ``zmin``      Minimum z value below which this annotation will not be
                  drawn.
    ``zmax``      Maximum z value above which this annotation will not be
                  drawn.
    ``creation``  Time of creation.
    ============= ============================================================

    All of these attributes can be modified directly, after which you should
    trigger a draw on the owning ``SliceCanvas`` to refresh the annotation.
    You shouldn't touch the ``expiry`` or ``creation`` attributes though.

    Subclasses must, at the very least, override the
    :meth:`globject.GLObject.draw2D` method.
    """


    enabled = props.Boolean(default=True)
    """Whether to draw this annotation or not. """


    lineWidth = props.Int(default=1, minval=0.1, clamped=True)
    """Line width, for annotations which are drawn with lines. """


    colour = props.Colour(default='#a00000')
    """Annotation colour."""


    alpha = props.Percentage(default=100)
    """Opacity."""


    honourZLimits = props.Boolean(default=False)
    """If True, the :attr:`zmin`/:attr:`zmax` properties are enforced.
    Otherwise (the default) they are ignored, and the annotation is
    always drawn.
    """


    zmin = props.Real()
    """Minimum z value below which this annotation will not be drawn. """


    zmax = props.Real()
    """Maximum z value below which this annotation will not be drawn. """


    def __init__(self,
                 annot,
                 xform=None,
                 colour=None,
                 alpha=None,
                 lineWidth=None,
                 enabled=True,
                 expiry=None,
                 honourZLimits=False,
                 zmin=None,
                 zmax=None,
                 **kwargs):
        """Create an ``AnnotationObject``.

        :arg annot:         The :class:`Annotations` object that created this
                            ``AnnotationObject``.

        :arg xform:         Transformation matrix which will be applied to all
                            vertex coordinates.

        :arg colour:        RGB/RGBA tuple specifying the annotation colour.

        :arg alpha:         Opacity.

        :arg lineWidth:     Line width to use for the annotation.

        :arg enabled:       Initially enabled or disabled.

        :arg expiry:        Time (in seconds) after which this annotation
                            should be expired and not drawn.

        :arg honourZLimits: Whether to enforce ``zmin``/``zmax``.

        :arg zmin:          Minimum z value below which this annotation should
                            not be drawn.

        :arg zmax:          Maximum z value above which this annotation should
                            not be drawn.

        Any other arguments are ignored.
        """
        globject.GLSimpleObject.__init__(self, False)

        self.annot    = annot
        self.xform    = xform
        self.creation = time.time()
        self.expiry   = expiry

        if colour        is not None: self.colour        = colour
        if alpha         is not None: self.alpha         = alpha
        if enabled       is not None: self.enabled       = enabled
        if lineWidth     is not None: self.lineWidth     = lineWidth
        if honourZLimits is not None: self.honourZLimits = honourZLimits
        if zmin          is not None: self.zmin          = zmin
        if zmax          is not None: self.zmax          = zmax

        if self.xform is not None:
            self.xform = np.array(self.xform, dtype=np.float32)


    def resetExpiry(self):
        """Resets the expiry for this ``AnnotationObject`` so that it is
        valid from the current time.
        """
        self.creation = time.time()


    def hit(self, x, y):
        """Return ``True`` if the given X/Y point is within the bounds of this
        annotation, ``False`` otherwise.  Must be implemented by sub-classes,
        but only those annotations which are drawn by the
        :class:`.OrthoAnnotateProfile`.

        :arg x: X coordinate (in display coordinates).
        :arg y: Y coordinate (in display coordinates).
        """
        raise NotImplementedError()


    def move(self, x, y):
        """Move this annotation acording to ``(x, y)``, which is specified as
        an offset relative to the current location. Must be implemented by
        sub-classes, but only those annotations which are drawn by the
        :class:`.OrthoAnnotateProfile`.

        :arg x: X coordinate (in display coordinates).
        :arg y: Y coordinate (in display coordinates).
        """
        raise NotImplementedError()


    def expired(self, now):
        """Returns ``True`` if this ``Annotation`` has expired, ``False``
        otherwise.

        :arg now: The current time
        """
        if self.expiry is None:
            return False

        return (self.creation + self.expiry) < now


    def preDraw(self, *args, **kwargs):
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)


    def postDraw(self, *args, **kwargs):
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)


class Point(AnnotationObject):
    """The ``Point`` class is an :class:`AnnotationObject` which represents a
    point, drawn as a small crosshair. The size of the point is proportional
    to the :attr:`AnnotationObject.lineWidth`.
    """


    def __init__(self, annot, x, y, *args, **kwargs):
        """Create a ``Point`` annotation.

        The ``xy`` coordinate tuple should be in relation to the axes which
        map to the horizontal/vertical screen axes on the target canvas.

        :arg annot: The :class:`Annotations` object that owns this ``Point``.

        :arg x:     X coordinates of the point

        :arg y:     Y coordinates of the point

        All other arguments are passed through to
        :meth:`AnnotationObject.__init__`.
        """
        AnnotationObject.__init__(self, annot, *args, **kwargs)
        self.x = x
        self.y = y


    def draw2D(self, zpos, axes):
        """Draws this ``Point`` annotation. """

        xax, yax, zax        = axes
        offset               = self.lineWidth * 0.5
        x, y                 = self.x, self.y
        idxs                 = np.arange(4,     dtype=np.uint32)
        verts                = np.zeros((4, 3), dtype=np.float32)
        verts[0, [xax, yax]] = [x - offset, y]
        verts[1, [xax, yax]] = [x + offset, y]
        verts[2, [xax, yax]] = [x, y - offset]
        verts[3, [xax, yax]] = [x, y + offset]
        verts[:, zax]        = zpos
        verts                = verts.ravel('C')

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_LINES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


    def hit(self, x, y):
        """Returns ``True`` if ``(x, y)`` is within the bounds of this
        ``Point``, ``False`` otherwise.
        """
        px, py = self.x, self.y
        dist   = np.sqrt((x - px) ** 2 + (y - py) ** 2)
        return dist <= (self.lineWidth * 0.5)


    def move(self, x, y):
        """Move this ``Point`` according to ``x`` and ``y``. """
        self.x = self.x + x
        self.y = self.y + y


class Line(AnnotationObject):
    """The ``Line`` class is an :class:`AnnotationObject` which represents a
    2D line.
    """


    def __init__(self, annot, x1, y1, x2, y2, *args, **kwargs):
        """Create a ``Line`` annotation.

        The ``xy1`` and ``xy2`` coordinate tuples should be in relation to the
        axes which map to the horizontal/vertical screen axes on the target
        canvas.

        :arg annot: The :class:`Annotations` object that owns this ``Line``.
        :arg x1:    X coordinate of one endpoint.
        :arg y1:    Y coordinate of one endpoint.
        :arg x2:    X coordinate of the other endpoint.
        :arg y2:    Y coordinate of the second endpoint.

        All other arguments are passed through to
        :meth:`AnnotationObject.__init__`.
        """
        AnnotationObject.__init__(self, annot, *args, **kwargs)
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2


    def draw2D(self, zpos, axes):
        """Draws this ``Line`` annotation. """

        xax, yax, zax = axes

        idxs                 = np.arange(2,     dtype=np.uint32)
        verts                = np.zeros((2, 3), dtype=np.float32)
        verts[0, [xax, yax]] = self.x1, self.y1
        verts[1, [xax, yax]] = self.x2, self.y2
        verts[:, zax]        = zpos
        verts                = verts.ravel('C')

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_LINES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


    def hit(self, x, y):
        """Returns ``True`` if ``(x, y)`` is within the bounds of this
        ``Line``, ``False`` otherwise.

        http://paulbourke.net/geometry/pointlineplane/
        """

        x1, y1 = self.x1, self.y1
        x2, y2 = self.x2, self.y2
        x3, y3 = x, y

        num = np.abs((x3 - x1) * (x2 - x1) + (y3 - y1) * (y2 - y1))
        dnm = (x2 - x1) ** 2 + (y2 - y1) ** 2
        u   = num / dnm

        ix   = x1 + u * (x2 - x1)
        iy   = y1 + u * (y2 - y1)
        dist = np.sqrt((x3 - ix) ** 2 + (y3 - iy) ** 2)

        # convert line width (in pixels) to display
        # coords, assume that pixels are isotropic
        thres = self.lineWidth * self.annot.canvas.pixelSize()[0]

        return dist <= (thres * 2)


    def move(self, x, y):
        """Move this ``Line`` according to ``(x, y)``."""
        self.x1 = self.x1 + x
        self.y1 = self.y1 + y
        self.x2 = self.x2 + x
        self.y2 = self.y2 + y


class Arrow(Line):
    """The ``Arrow`` class is an :class:`AnnotationObject` which represents a
    2D line with an arrow head at one end. The size of the is proportional
    to the current :attr:`AnnotationObject.lineWidth`.
    """


    def draw2D(self, zpos, axes):
        """Draw the arrow. """

        Line.draw2D(self, zpos, axes)

        xax, yax, zax = axes

        # We draw the arrow head as a triangle at the
        # second line vertex (xy2). We generate the
        # two other vertices of the triangle by
        # rotating +/- 30 degrees around xy2.
        xy1   = np.array([self.x1, self.y1])
        xy2   = np.array([self.x2, self.y2])
        vec   = xy2 - xy1
        vec   = vec / np.sqrt(vec[0] ** 2 + vec[1] ** 2)
        angle = np.arccos(np.dot(vec, [1, 0])) * np.sign(vec[1])
        delta = np.pi / 6

        p1 = np.array((np.cos(angle + delta), np.sin(angle + delta)))
        p2 = np.array((np.cos(angle - delta), np.sin(angle - delta)))

        # We also add a little padding to xy2 because
        # otherwise the main line may appear beyond
        # the triangle if a large line width is set.
        xy2 = xy2 + self.lineWidth * vec * 0.5
        p1  = xy2 - self.lineWidth * p1
        p2  = xy2 - self.lineWidth * p2

        idxs  = np.arange(3, dtype=np.uint32)
        verts = np.zeros((3, 3), dtype=np.float32)
        verts[0, [xax, yax]] = xy2
        verts[1, [xax, yax]] = p1
        verts[2, [xax, yax]] = p2
        verts[:, zax]        = zpos

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_TRIANGLES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


class Rect(AnnotationObject):
    """The ``Rect`` class is an :class:`AnnotationObject` which represents a
    2D rectangle.
    """


    filled = props.Boolean(default=True)
    """Whether to fill the rectangle. """


    border = props.Boolean(default=True)
    """Whether to draw a border around the rectangle. """


    def __init__(self,
                 annot,
                 x,
                 y,
                 w,
                 h,
                 filled=True,
                 border=True,
                 *args,
                 **kwargs):
        """Create a :class:`Rect` annotation.

        :arg annot:  The :class:`Annotations` object that owns this
                     ``Rect``.

        :arg x:      X coordinate of one  corner of the rectangle, in
                     the display coordinate system.

        :arg y:      Y coordinate of one  corner of the rectangle, in
                     the display coordinate system.

        :arg w:      Rectangle width (actually an offset relative to ``x``)

        :arg h:      Rectangle height (actually an offset relative to ``y``)

        :arg filled: If ``True``, the rectangle is filled

        :arg border: If ``True``, a border is drawn around the rectangle.

        All other arguments are passed through to
        :meth:`AnnotationObject.__init__`.

        Note that if ``filled=False`` and ``border=False``, nothing will be
        drawn. The ``.AnnotationObject.alpha`` value is ignored when drawing
        the border.
        """
        AnnotationObject.__init__(self, annot, *args, **kwargs)

        self.x      = x
        self.y      = y
        self.w      = w
        self.h      = h
        self.filled = filled
        self.border = border


    def hit(self, x, y):
        """Returns ``True`` if ``(x, y)`` is within the bounds of this
        ``Rect``, ``False`` otherwise.
        """

        xlo, ylo = self.x, self.y
        xhi, yhi = (xlo + self.w, ylo + self.h)

        xlo, xhi = sorted((xlo, xhi))
        ylo, yhi = sorted((ylo, yhi))

        return (x >= xlo and
                x <= xhi and
                y >= ylo and
                y <= yhi)


    def move(self, x, y):
        """Move this ``Rect`` according to ``(x, y)``."""
        self.x = self.x + x
        self.y = self.y + y


    def draw2D(self, zpos, axes):
        """Draws this ``Rectangle`` annotation. """

        if self.w == 0 or self.h == 0:
            return

        xax, yax, zax = axes
        x, y          = self.x, self.y
        w             = self.w
        h             = self.h

        bl = [x,     y]
        br = [x + w, y]
        tl = [x,     y + h]
        tr = [x + w, y + h]

        if self.border:
            self.__drawRect(zpos, xax, yax, zax, bl, br, tl, tr)

        if self.filled:
            self.__drawFill(zpos, xax, yax, zax, bl, br, tl, tr)


    def __drawFill(self, zpos, xax, yax, zax, bl, br, tl, tr):
        """Draw a filled version of the rectangle. """

        if self.colour is not None: colour = list(self.colour[:3])
        else:                       colour = [1, 1, 1]

        colour = colour + [self.alpha / 100]

        idxs  = np.array([0, 1, 2, 2, 1, 3], dtype=np.uint32)
        verts = np.zeros((4, 3),             dtype=np.float32)

        verts[0, [xax, yax]] = bl
        verts[1, [xax, yax]] = br
        verts[2, [xax, yax]] = tl
        verts[3, [xax, yax]] = tr
        verts[:,  zax]       = zpos
        verts                = verts.ravel('C')

        # I'm assuming that glPolygonMode
        # is already set to GL_FILL
        gl.glColor4f(*colour)
        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_TRIANGLES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


    def __drawRect(self, zpos, xax, yax, zax, bl, br, tl, tr):
        """Draw the rectangle outline. """

        if self.colour is not None: colour = list(self.colour[:3]) + [1]
        else:                       colour = [1, 1, 1, 1]

        idxs  = np.array([0, 1, 2, 3, 0, 2, 1, 3], dtype=np.uint32)
        verts = np.zeros((4, 3),                   dtype=np.float32)

        verts[0, [xax, yax]] = bl
        verts[1, [xax, yax]] = br
        verts[2, [xax, yax]] = tl
        verts[3, [xax, yax]] = tr
        verts[:,  zax]       = zpos
        verts                = verts.ravel('C')

        gl.glColor4f(*colour)

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_LINES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


class Ellipse(AnnotationObject):
    """The ``Ellipse`` class is an :class:`AnnotationObject` which represents a
    ellipse.
    """


    filled = props.Boolean(default=True)
    """Whether to fill the ellipse. """


    border = props.Boolean(default=True)
    """Whether to draw a border around the ellipse. """


    def __init__(self,
                 annot,
                 x,
                 y,
                 w,
                 h,
                 npoints=60,
                 filled=True,
                 border=True,
                 *args, **kwargs):
        """Create an ``Ellipse`` annotation.

        :arg annot:   The :class:`Annotations` object that owns this
                      ``Ellipse``.

        :arg x:       X coordinate of ellipse centre, in the display
                      coordinate system.

        :arg y:       Y coordinate of ellipse centre, in the display
                      coordinate system.

        :arg w:       Horizontal radius.

        :arg h:       Vertical radius.

        :arg npoints: Number of vertices used to draw the ellipse outline.

        :arg filled:  If ``True``, the ellipse is filled

        :arg border: If ``True``, a border is drawn around the ellipse

        All other arguments are passed through to
        :meth:`AnnotationObject.__init__`.
        """

        AnnotationObject.__init__(self, annot, *args, **kwargs)

        self.x       = x
        self.y       = y
        self.w       = w
        self.h       = h
        self.npoints = npoints
        self.filled  = filled
        self.border  = border


    def hit(self, x, y):
        """Returns ``True`` if ``(x, y)`` is within the bounds of this
        ``Ellipse``, ``False`` otherwise.
        """

        # https://math.stackexchange.com/a/76463
        h,  k  = self.x, self.y
        rx, ry = self.w, self.h
        return ((x - h) ** 2) / (rx ** 2) + ((y - k) ** 2) / (ry ** 2) <= 1


    def move(self, x, y):
        """Move this ``Rect`` according to ``(x, y)``."""
        self.x = self.x + x
        self.y = self.y + y


    def draw2D(self, zpos, axes):
        """Draws this ``Ellipse`` annotation. """

        if (self.w == 0) or (self.h == 0):
            return

        if self.colour is not None: colour = list(self.colour[:3])
        else:                       colour = [1, 1, 1]

        r, g, b = colour
        a       = self.alpha / 100

        xax, yax, zax = axes
        x, y          = self.x, self.y
        w, h          = self.w, self.h

        idxs    = np.arange(self.npoints + 1, dtype=np.uint32)
        verts   = np.zeros((self.npoints + 1, 3), dtype=np.float32)
        samples = np.linspace(0, 2 * np.pi, self.npoints)

        verts[0, [xax, yax]] = x, y
        verts[1:, xax]       = w * np.sin(samples) + x
        verts[1:, yax]       = h * np.cos(samples) + y
        verts[:,  zax]       = zpos

        # border
        if self.border:
            gl.glColor4f(r, g, b, 1)
            gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts[1:-1])
            gl.glDrawElements(
                gl.GL_LINE_LOOP, len(idxs) - 2, gl.GL_UNSIGNED_INT, idxs[:-2])

        if self.filled:
            gl.glColor4f(r, g, b, a)
            gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
            gl.glDrawElements(
                gl.GL_TRIANGLE_FAN, len(idxs), gl.GL_UNSIGNED_INT, idxs)


class VoxelSelection(AnnotationObject):
    """A ``VoxelSelection`` is an :class:`AnnotationObject` which draws
    selected voxels from a :class:`.selection.Selection` instance.  A
    :class:`.SelectionTexture` is used to draw the selected voxels.
    """


    def __init__(self,
                 annot,
                 selection,
                 opts,
                 offsets=None,
                 *args,
                 **kwargs):
        """Create a ``VoxelSelection`` annotation.

        :arg annot:     The :class:`Annotations` object that owns this
                        ``VoxelSelection``.

        :arg selection: A :class:`.selection.Selection` instance which defines
                        the voxels to be highlighted.

        :arg opts:      A :class:`.NiftiOpts` instance which is used
                        for its voxel-to-display transformation matrices.

        :arg offsets:   If ``None`` (the default), the ``selection`` must have
                        the same shape as the image data being
                        annotated. Alternately, you may set ``offsets`` to a
                        sequence of three values, which are used as offsets
                        for the xyz voxel values. This is to allow for a
                        sub-space of the full image space to be annotated.

        All other arguments are passed through to the
        :meth:`AnnotationObject.__init__` method.
        """

        AnnotationObject.__init__(self, annot, *args, **kwargs)

        if offsets is None:
            offsets = [0, 0, 0]

        self.__selection = selection
        self.__opts      = opts
        self.__offsets   = offsets

        texName = '{}_{}'.format(type(self).__name__, id(selection))

        ndims = texdata.numTextureDims(selection.shape)

        if ndims == 2: ttype = textures.SelectionTexture2D
        else:          ttype = textures.SelectionTexture3D

        self.__texture = glresources.get(
            texName,
            ttype,
            texName,
            selection)


    def destroy(self):
        """Must be called when this ``VoxelSelection`` is no longer needed.
        Destroys the :class:`.SelectionTexture`.
        """
        glresources.delete(self.__texture.name)
        self.__texture = None
        self.__opts    = None


    @property
    def texture(self):
        """Return the :class:`.SelectionTexture` used by this
        ``VoxelSelection``.
        """
        return self.__texture


    def draw2D(self, zpos, axes):
        """Draws this ``VoxelSelection``."""

        xax, yax     = axes[:2]
        opts         = self.__opts
        texture      = self.__texture
        shape        = self.__selection.getSelection().shape
        displayToVox = opts.getTransform('display', 'voxel')
        voxToDisplay = opts.getTransform('voxel',   'display')
        voxToTex     = opts.getTransform('voxel',   'texture')
        voxToTex     = affine.concat(texture.texCoordXform(shape), voxToTex)
        verts, voxs  = glroutines.slice2D(shape,
                                          xax,
                                          yax,
                                          zpos,
                                          voxToDisplay,
                                          displayToVox)

        texs  = affine.transform(voxs, voxToTex)[:, :texture.ndim]
        verts = np.array(verts, dtype=np.float32).ravel('C')
        texs  = np.array(texs,  dtype=np.float32).ravel('C')

        texture.bindTexture(gl.GL_TEXTURE0)
        gl.glClientActiveTexture(gl.GL_TEXTURE0)
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_MODULATE)

        with glroutines.enabled((texture.target,
                                 gl.GL_TEXTURE_COORD_ARRAY,
                                 gl.GL_VERTEX_ARRAY)):
            gl.glVertexPointer(  3,            gl.GL_FLOAT, 0, verts)
            gl.glTexCoordPointer(texture.ndim, gl.GL_FLOAT, 0, texs)
            gl.glDrawArrays(     gl.GL_TRIANGLES, 0, 6)

        texture.unbindTexture()


class TextAnnotation(AnnotationObject):
    """A ``TextAnnotation`` is an ``AnnotationObject`` which draws a
    :class:`fsleyes.gl.text.Text` object.

    The ``Text`` class allows the text position to be specified as either
    x/y proportions, or as absolute pixels. The ``TextAnnotation`` class
    adds an additional option to specify the location in terms of a
    3D position in the display coordinate system.

    This can be achieved by setting ``coordinates`` to ``'display'``, and
    setting ``pos`` to the 3D position in the display coordinate system.
    """


    text  = props.String()
    """Text to draw. """


    fontSize = props.Int(minval=6, maxval=48)
    """Text font size in points. The size of the text annotation is kept
    proportional to the canvas zoom level.
    """


    def __init__(self,
                 annot,
                 text=None,
                 x=None,
                 y=None,
                 off=None,
                 coordinates='proportions',
                 fontSize=10,
                 halign=None,
                 valign=None,
                 colour=None,
                 **kwargs):
        """Create a ``TextAnnotation``.

        :arg annot:      The :class:`Annotations` object that owns this
                         ``TextAnnotation``.

        See the :class:`.Text` class for details on the other arguments.
        """
        AnnotationObject.__init__(self, annot, **kwargs)

        self.text        = text
        self.x           = x
        self.y           = y
        self.off         = off
        self.coordinates = coordinates
        self.fontSize    = fontSize
        self.halign      = halign
        self.valign      = valign
        self.colour      = colour
        self.__initscale = None
        self.__text      = gltext.Text()


    @property
    def gltext(self):
        return self.__text


    def destroy(self):
        """Must be called when this ``TextAnnotation`` is no longer needed.
        """
        AnnotationObject.destroy(self)
        self.__text.destroy()
        self.__text = None


    def draw2D(self, zpos, axes):
        """Draw this ``TextAnnotation``. """

        if self.colour is not None: colour = self.colour[:3]
        else:                       colour = [1, 1, 1]

        text             = self.__text
        canvas           = self.annot.canvas
        opts             = canvas.opts
        text.text        = self.text
        text.off         = self.off
        text.coordinates = self.coordinates
        text.fontSize    = self.fontSize
        text.halign      = self.halign
        text.valign      = self.valign
        text.colour      = colour
        text.alpha       = self.alpha / 100

        if self.coordinates == 'display':

            # Make sure the text is sized proportional
            # to the display coordinate system, so
            # invariant to zoom factor/canvas size
            if self.__initscale is None:
                w, h             = canvas.pixelSize()
                self.__initscale = np.sqrt(w * h)

            w, h  = canvas.pixelSize()
            scale = np.sqrt(w * h)

            pos           = [0] * 3
            pos[opts.xax] = self.x
            pos[opts.yax] = self.y
            pos[opts.zax] = opts.pos[2]

            text.pos         = canvas.worldToCanvas(pos)
            text.scale       = self.__initscale / scale
            text.coordinates = 'pixels'
        else:
            text.pos         = self.x, self.y
            text.coordinates = self.coordinates

        text.draw(*canvas.GetSize())


    def hit(self, x, y):
        """Returns ``True`` if ``(x, y)`` is within the bounds of this
        ``TextAnnotation``, ``False`` otherwise. Only supported for text
        drawn relative to the display coordinate system
        (``coordinates='display'``) - raises a ``NotImplementedError``
        otherwise.
        """

        if self.coordinates != 'display':
            raise NotImplementedError()

        canvas     = self.annot.canvas
        opts       = canvas.opts
        xlo,  ylo  = self.__text.pos
        xlen, ylen = self.__text.size

        # the Text object works in pixels, but
        # here we're working in display coords
        xy1 = canvas.canvasToWorld(xlo,        ylo)
        xy2 = canvas.canvasToWorld(xlo + xlen, ylo + ylen)

        xlo, xhi = sorted((xy1[opts.xax], xy2[opts.xax]))
        ylo, yhi = sorted((xy1[opts.yax], xy2[opts.yax]))

        return (x >= xlo and
                x <= xhi and
                y >= ylo and
                y <= yhi)


    def move(self, x, y):
        """Move this ``TextAnnotation`` according to ``(x, y)``.

        Only supported for text drawn relative to the display coordinate
        system (``coordinates='display'``) - raises a ``NotImplementedError``
        otherwise.
        """
        if self.coordinates != 'display':
            raise NotImplementedError()

        self.x = self.x + x
        self.y = self.y + y
